# Copyright 2021 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from typing import List, Union, Optional, get_type_hints

from pydantic import parse_obj_as

from dwave.cloud.api.client import DWaveAPIClient, SolverAPIClient, MetadataAPIClient
from dwave.cloud.api import constants, models
from dwave.cloud.utils import NumpyEncoder

__all__ = ['Solvers', 'Problems', 'Regions']


class ResourceBase:
    """A class for interacting with a SAPI resource."""

    resource_path = None

    def _patch_session(self):
        # anchor all session requests at the new base path
        if self.resource_path and self.session:
            self.session.base_url = self.session.create_url(self.resource_path)

    def __init__(self, **config):
        self.client = DWaveAPIClient(**config)
        self.session = self.client.session
        self._patch_session()

    def close(self):
        self.client.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    @classmethod
    def from_client_config(cls, client: Union[DWaveAPIClient, 'dwave.cloud.client.base.Client']):
        """Create Resource instance configured from a
        :class:`~dwave.cloud.client.base.Client' instance.
        """
        # TODO: also accept configuration dict/dataclass when config/client refactored
        if isinstance(client, DWaveAPIClient):
            return cls(**client.config)
        else: # assume isinstance(client, dwave.cloud.Client), without importing
            sapiclient = SolverAPIClient.from_client_config(client)
            return cls(**sapiclient.config)


class Regions(ResourceBase):

    resource_path = 'regions/'

    def __init__(self, **config):
        self.client = MetadataAPIClient(**config)
        self.session = self.client.session
        self._patch_session()

    # Content-Type: application/vnd.dwave.metadata.regions+json; version=1.0.0
    def list_regions(self) -> List[models.Region]:
        path = ''
        response = self.session.get(path)
        regions = response.json()
        return parse_obj_as(List[models.Region], regions)

    # Content-Type: application/vnd.dwave.metadata.region+json; version=1.0.0
    def get_region(self, code: str) -> models.Region:
        path = '{}'.format(code)
        response = self.session.get(path)
        region = response.json()
        return parse_obj_as(models.Region, region)


class Solvers(ResourceBase):

    resource_path = 'solvers/'

    # Content-Type: application/vnd.dwave.sapi.solver-definition-list+json; version=2.0.0
    def list_solvers(self) -> List[models.SolverConfiguration]:
        path = 'remote/'
        response = self.session.get(path)
        solvers = response.json()
        return parse_obj_as(List[models.SolverConfiguration], solvers)

    # Content-Type: application/vnd.dwave.sapi.solver-definition+json; version=2.0.0
    def get_solver(self, solver_id: str) -> models.SolverConfiguration:
        path = 'remote/{}'.format(solver_id)
        response = self.session.get(path)
        solver = response.json()
        return models.SolverConfiguration.parse_obj(solver)


class Problems(ResourceBase):

    resource_path = 'problems/'

    # Content-Type: application/vnd.dwave.sapi.problems+json; version=2.1.0
    def list_problems(self, *,
                      id: str = None,
                      label: str = None,
                      max_results: int = None,
                      status: Union[str, constants.ProblemStatus] = None,
                      solver: str = None,
                      **params) -> List[models.ProblemStatus]:
        """Retrieve a filtered list of submitted problems."""

        # build query
        if id is not None:
            params.setdefault('id', id)
        if label is not None:
            params.setdefault('label', label)
        if max_results is not None:
            params.setdefault('max_results', max_results)
        if isinstance(status, constants.ProblemStatus):
            params.setdefault('status', status.value)
        elif status is not None:
            params.setdefault('status', status)
        if solver is not None:
            params.setdefault('solver', solver)

        path = ''
        response = self.session.get(path, params=params)
        statuses = response.json()
        return parse_obj_as(List[models.ProblemStatus], statuses)

    # Content-Type: application/vnd.dwave.sapi.problem+json; version=2.1.0
    def get_problem(self, problem_id: str) -> models.ProblemStatusMaybeWithAnswer:
        """Retrieve problem short status and answer if answer is available."""
        path = '{}'.format(problem_id)
        response = self.session.get(path)
        status = response.json()
        return models.ProblemStatusMaybeWithAnswer.parse_obj(status)

    # Content-Type: application/vnd.dwave.sapi.problems+json; version=2.1.0
    def get_problem_status(self, problem_id: str) -> models.ProblemStatus:
        """Retrieve short status of a single problem."""
        path = ''
        params = dict(id=problem_id)
        response = self.session.get(path, params=params)
        status = response.json()[0]
        return models.ProblemStatus.parse_obj(status)

    # Content-Type: application/vnd.dwave.sapi.problems+json; version=2.1.0
    # XXX: @pydantic.validate_arguments
    def get_problem_statuses(self, problem_ids: List[str]) -> List[models.ProblemStatus]:
        """Retrieve short problem statuses for a list of problems."""
        if not isinstance(problem_ids, list):
            raise TypeError('a list of problem ids expected')
        if len(problem_ids) > 1000:
            raise ValueError('number of problem ids is limited to 1000')

        path = ''
        params = dict(id=','.join(problem_ids))
        response = self.session.get(path, params=params)
        statuses = response.json()
        return parse_obj_as(List[models.ProblemStatus], statuses)

    # Content-Type: application/vnd.dwave.sapi.problem-data+json; version=2.1.0
    def get_problem_info(self, problem_id: str) -> models.ProblemInfo:
        """Retrieve complete problem info."""
        path = '{}/info'.format(problem_id)
        response = self.session.get(path)
        info = response.json()
        return models.ProblemInfo.parse_obj(info)

    # Content-Type: application/vnd.dwave.sapi.problem-answer+json; version=2.1.0
    def get_problem_answer(self, problem_id: str) -> models.ProblemAnswer:
        """Retrieve problem answer."""
        path = '{}/answer'.format(problem_id)
        response = self.session.get(path)
        answer = response.json()['answer']
        return models.ProblemAnswer.parse_obj(answer)

    # Content-Type: application/vnd.dwave.sapi.problem-message+json; version=2.1.0
    def get_problem_messages(self, problem_id: str) -> List[dict]:
        """Retrieve list of problem messages."""
        path = '{}/messages'.format(problem_id)
        response = self.session.get(path)
        return response.json()

    # Content-Type: application/vnd.dwave.sapi.problems+json; version=2.1.0
    def submit_problem(self, *,
                       data: models.ProblemData,
                       params: dict,
                       solver: str,
                       type: constants.ProblemType,
                       label: str = None) -> \
            Union[models.ProblemStatusMaybeWithAnswer, models.ProblemSubmitError]:
        """Blocking problem submit with timeout, returning final status and
        answer, if problem is solved within the (undisclosed) time limit.
        """
        path = ''
        body = dict(data=data.dict(), params=params, solver=solver,
                    type=type, label=label)
        data = json.dumps(body, cls=NumpyEncoder)
        response = self.session.post(
            path, data=data, headers={'Content-Type': 'application/json'})
        rtype = get_type_hints(self.submit_problem)['return']
        return parse_obj_as(rtype, response.json())

    # Content-Type: application/vnd.dwave.sapi.problems+json; version=2.1.0
    def submit_problems(self, problems: List[models.ProblemJob]) -> \
            List[Union[models.ProblemInitialStatus, models.ProblemSubmitError]]:
        """Asynchronous multi-problem submit, returning initial statuses."""
        path = ''
        # encode iteratively so that timestamps are serialized (via pydantic json encoder)
        body = '[%s]' % ','.join(p.json() for p in problems)
        response = self.session.post(
            path, data=body, headers={'Content-Type': 'application/json'})
        rtype = get_type_hints(self.submit_problems)['return']
        return parse_obj_as(rtype, response.json())

    # Content-Type: application/vnd.dwave.sapi.problem+json; version=2.1.0
    def cancel_problem(self, problem_id: str) -> \
            Union[models.ProblemStatus, models.ProblemCancelError]:
        """Initiate problem cancel by problem id."""
        path = '{}'.format(problem_id)
        response = self.session.delete(path)
        rtype = get_type_hints(self.cancel_problem)['return']
        return parse_obj_as(rtype, response.json())

    # Content-Type: application/vnd.dwave.sapi.problems+json; version=2.1.0
    def cancel_problems(self, problem_ids: List[str]) -> \
            List[Union[models.ProblemStatus, models.ProblemCancelError]]:
        """Initiate problem cancel for a list of problems."""
        path = ''
        response = self.session.delete(path, json=problem_ids)
        rtype = get_type_hints(self.cancel_problems)['return']
        return parse_obj_as(rtype, response.json())



# Copyright 2017 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from dwave.cloud.qpu import Client


# setup local logger
formatter = logging.Formatter('%(asctime)s %(name)s %(levelname)s %(message)s')
handler = logging.StreamHandler()
handler.setFormatter(formatter)
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
logger.addHandler(handler)


with Client.from_config(profile='prod') as client:
    solvers = client.get_solvers()
    logger.info("Solvers available: %r", solvers)

    solver = client.get_solver()
    comp = solver.sample_qubo({})

    logger.info("Problem submitted:")
    logger.info(" - time received: %s", comp.time_received)
    logger.info(" - time solved: %s", comp.time_solved)
    logger.info(" - parse time: %s", comp.parse_time)
    logger.info(" - remote status: %s", comp.remote_status)
    logger.info(" - min_eta: %s", comp.eta_min)

    result = comp.result()
    logger.info("Result received:")
    logger.info(" - time received: %s", comp.time_received)
    logger.info(" - time solved: %s", comp.time_solved)
    logger.info(" - parse time: %s", comp.parse_time)
    logger.info(" - remote status: %s", comp.remote_status)



# Copyright 2017 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import json
import uuid
import logging
import unittest
import warnings
import tempfile
from unittest import mock
from collections import OrderedDict
from itertools import count
from datetime import datetime
from functools import partial

import numpy
from parameterized import parameterized

from dwave.cloud import FilteredSecretsFormatter
from dwave.cloud.utils import (
    uniform_iterator, uniform_get, strip_head, strip_tail,
    active_qubits, generate_random_ising_problem, NumpyEncoder,
    default_text_input, utcnow, cached, retried, deprecated, aliasdict,
    parse_loglevel, user_agent, hasinstance, exception_chain, is_caused_by)


class TestSimpleUtils(unittest.TestCase):

    def test_uniform_iterator(self):
        items = [('a', 1), ('b', 2)]
        self.assertEqual(list(uniform_iterator(OrderedDict(items))), items)
        self.assertEqual(list(uniform_iterator('ab')), list(enumerate('ab')))

    def test_uniform_get(self):
        d = {0: 0, 1: 1}
        self.assertEqual(uniform_get(d, 0), 0)
        self.assertEqual(uniform_get(d, 2), None)
        self.assertEqual(uniform_get(d, 2, default=0), 0)
        l = [0, 1]
        self.assertEqual(uniform_get(l, 0), 0)
        self.assertEqual(uniform_get(l, 2), None)
        self.assertEqual(uniform_get(l, 2, default=0), 0)

    def test_strip_head(self):
        self.assertEqual(strip_head([0, 0, 1, 2], [0]), [1, 2])
        self.assertEqual(strip_head([1], [0]), [1])
        self.assertEqual(strip_head([1], []), [1])
        self.assertEqual(strip_head([0, 0, 1, 2], [0, 1, 2]), [])

    def test_strip_tail(self):
        self.assertEqual(strip_tail([1, 2, 0, 0], [0]), [1, 2])
        self.assertEqual(strip_tail([1], [0]), [1])
        self.assertEqual(strip_tail([1], []), [1])
        self.assertEqual(strip_tail([0, 0, 1, 2], [0, 1, 2]), [])

    def test_active_qubits_dict(self):
        self.assertEqual(active_qubits({}, {}), set())
        self.assertEqual(active_qubits({0: 0}, {}), {0})
        self.assertEqual(active_qubits({}, {(0, 1): 0}), {0, 1})
        self.assertEqual(active_qubits({2: 0}, {(0, 1): 0}), {0, 1, 2})

    def test_active_qubits_list(self):
        self.assertEqual(active_qubits([], {}), set())
        self.assertEqual(active_qubits([2], {}), {0})
        self.assertEqual(active_qubits([2, 2, 0], {}), {0, 1, 2})
        self.assertEqual(active_qubits([], {(0, 1): 0}), {0, 1})
        self.assertEqual(active_qubits([0, 0], {(0, 2): 0}), {0, 1, 2})

    def test_default_text_input(self):
        val = "value"
        with mock.patch("click.termui.visible_prompt_func", side_effect=[val]):
            self.assertEqual(default_text_input("prompt", val), val)
        with mock.patch("click.termui.visible_prompt_func", side_effect=[val]):
            self.assertEqual(default_text_input("prompt", val+val), val)

    def test_optional_text_input(self):
        with mock.patch("click.termui.visible_prompt_func", side_effect=[""]):
            self.assertEqual(default_text_input("prompt", optional=True), None)

    def test_optional_choices_text_input(self):
        with mock.patch("click.termui.visible_prompt_func", side_effect=[""]):
            self.assertEqual(
                default_text_input("prompt", choices='abc', optional=True), None)
        with mock.patch("click.termui.visible_prompt_func", side_effect=["d", "skip"]):
            self.assertEqual(
                default_text_input("prompt", choices='abc', optional=True), None)
        with mock.patch("click.termui.visible_prompt_func", side_effect=["e", "a"]):
            self.assertEqual(
                default_text_input("prompt", choices='abc', optional=True), 'a')

    def test_generate_random_ising_problem(self):
        class MockSolver(object):
            nodes = [0, 1, 3]
            undirected_edges = {(0, 1), (1, 3), (0, 4)}
            properties = dict(h_range=[2, 2], j_range=[-1, -1])
        mock_solver = MockSolver()

        lin, quad = generate_random_ising_problem(mock_solver)

        self.assertDictEqual(lin, {0: 2.0, 1: 2.0, 3: 2.0})
        self.assertDictEqual(quad, {(0, 1): -1.0, (1, 3): -1.0, (0, 4): -1.0})

    def test_generate_random_ising_problem_default_solver_ranges(self):
        class MockSolver(object):
            nodes = [0, 1, 3]
            undirected_edges = {(0, 1), (1, 3), (0, 4)}
            properties = {}
        mock_solver = MockSolver()

        lin, quad = generate_random_ising_problem(mock_solver)

        for q, v in lin.items():
            self.assertTrue(v >= -1 and v <= 1)
        for e, v in quad.items():
            self.assertTrue(v >= -1 and v <= 1)

    def test_generate_random_ising_problem_with_user_constrained_ranges(self):
        class MockSolver(object):
            nodes = [0, 1, 3]
            undirected_edges = {(0, 1), (1, 3), (0, 4)}
            properties = dict(h_range=[2, 2], j_range=[-1, -1])
        mock_solver = MockSolver()

        lin, quad = generate_random_ising_problem(mock_solver, h_range=[0,0], j_range=[1,1])

        self.assertDictEqual(lin, {0: 0.0, 1: 0.0, 3: 0.0})
        self.assertDictEqual(quad, {(0, 1): 1.0, (1, 3): 1.0, (0, 4): 1.0})

    def test_utcnow(self):
        t = utcnow()
        now = datetime.utcnow()
        self.assertEqual(t.utcoffset().total_seconds(), 0.0)
        unaware = t.replace(tzinfo=None)
        self.assertLess((now - unaware).total_seconds(), 1.0)

    def test_parse_loglevel_invalid(self):
        """Parsing invalid log levels returns NOTSET."""
        notset = logging.NOTSET

        self.assertEqual(parse_loglevel(''), notset)
        self.assertEqual(parse_loglevel('  '), notset)
        self.assertEqual(parse_loglevel(None), notset)
        self.assertEqual(parse_loglevel(notset), notset)
        self.assertEqual(parse_loglevel('nonexisting'), notset)
        self.assertEqual(parse_loglevel({'a': 1}), notset)
        self.assertIsNone(parse_loglevel('nonexisting', default=None))

    def test_parse_loglevel_numeric_and_symbolic(self):
        self.assertEqual(parse_loglevel('info'), logging.INFO)
        self.assertEqual(parse_loglevel('INFO'), logging.INFO)
        self.assertEqual(parse_loglevel(logging.INFO), logging.INFO)
        self.assertEqual(parse_loglevel(str(logging.INFO)), logging.INFO)
        self.assertEqual(parse_loglevel('  %d  ' % logging.INFO), logging.INFO)

    def test_user_agent(self):
        from dwave.cloud.package_info import __packagename__, __version__
        ua = user_agent(__packagename__, __version__)

        required = [__packagename__, 'python', 'machine', 'system', 'platform']
        for key in required:
            self.assertIn(key, ua)


# copied from dwave-hybrid utils
# TODO: remove these tests when we/if switch to `dwave.common`
class TestNumpyJSONEncoder(unittest.TestCase):

    @parameterized.expand([
        (numpy.bool_(1), True), (numpy.bool8(1), True),
        (numpy.byte(1), 1), (numpy.int8(1), 1),
        (numpy.ubyte(1), 1), (numpy.uint8(1), 1),
        (numpy.short(1), 1), (numpy.int16(1), 1),
        (numpy.ushort(1), 1), (numpy.uint16(1), 1),
        (numpy.intc(1), 1), (numpy.int32(1), 1),
        (numpy.uintc(1), 1), (numpy.uint32(1), 1),
        (numpy.int_(1), 1), (numpy.int32(1), 1),
        (numpy.uint(1), 1), (numpy.uint32(1), 1),
        (numpy.longlong(1), 1), (numpy.int64(1), 1),
        (numpy.ulonglong(1), 1), (numpy.uint64(1), 1),
        (numpy.half(1.0), 1.0), (numpy.float16(1.0), 1.0),
        (numpy.single(1.0), 1.0), (numpy.float32(1.0), 1.0),
        (numpy.double(1.0), 1.0), (numpy.float64(1.0), 1.0),
        (numpy.longdouble(1.0), 1.0)
    ] + ([
        (numpy.float128(1.0), 1.0)      # unavailable on windows
    ] if hasattr(numpy, 'float128') else [
    ]))
    def test_numpy_primary_type_encode(self, np_val, py_val):
        self.assertEqual(
            json.dumps(py_val),
            json.dumps(np_val, cls=NumpyEncoder)
        )

    @parameterized.expand([
        (numpy.array([1, 2, 3], dtype=numpy.int), [1, 2, 3]),
        (numpy.array([[1], [2], [3]], dtype=numpy.double), [[1.0], [2.0], [3.0]]),
        (numpy.zeros((2, 2), dtype=numpy.bool_), [[False, False], [False, False]]),
        (numpy.array([('Rex', 9, 81.0), ('Fido', 3, 27.0)],
                     dtype=[('name', 'U10'), ('age', 'i4'), ('weight', 'f4')]),
         [['Rex', 9, 81.0], ['Fido', 3, 27.0]]),
        (numpy.rec.array([(1, 2., 'Hello'), (2, 3., "World")],
                         dtype=[('foo', 'i4'), ('bar', 'f4'), ('baz', 'U10')]),
         [[1, 2.0, "Hello"], [2, 3.0, "World"]])
    ])
    def test_numpy_array_encode(self, np_val, py_val):
        self.assertEqual(
            json.dumps(py_val),
            json.dumps(np_val, cls=NumpyEncoder)
        )


class TestCachedInMemoryDecorator(unittest.TestCase):
    """Test @cached using in-memory store."""

    def setUp(self):
        self.cached = cached

    def test_args_hashing(self):
        counter = count()

        @self.cached(maxage=300)
        def f(*a, **b):
            return next(counter)

        with mock.patch('dwave.cloud.utils.epochnow', lambda: 0):
            self.assertEqual(f(), 0)
            self.assertEqual(f(1), 1)
            self.assertEqual(f(1, 2), 2)
            self.assertEqual(f(1), 1)
            self.assertEqual(f(1, refresh_=True), 3)
            self.assertEqual(f(1, 2), 2)

            self.assertEqual(f(a=1, b=2), 4)
            self.assertEqual(f(b=2, a=1), 4)
            self.assertEqual(f(b=2, a=1, refresh_=1), 5)
            self.assertEqual(f(), 0)

            self.assertEqual(f(2), 6)
            self.assertEqual(f(1), 3)

    def test_args_collision(self):
        counter = count()

        @self.cached(maxage=300)
        def f(*a, **b):
            return next(counter)

        with mock.patch('dwave.cloud.utils.epochnow', lambda: 0):
            # NB: in python2, without hash seed randomization,
            # hash('\0B') == hash('\0\0C')
            self.assertEqual(f(x='\0B'), 0)
            self.assertEqual(f(x='\0\0C'), 1)

    def test_expiry(self):
        counter = count()

        @self.cached(maxage=300)
        def f(*a, **b):
            return next(counter)

        # populate
        with mock.patch('dwave.cloud.utils.epochnow', lambda: 0):
            self.assertEqual(f(), 0)
            self.assertEqual(f(1), 1)
            self.assertEqual(f(a=1, b=2), 2)

        # verify expiry
        with mock.patch('dwave.cloud.utils.epochnow', lambda: 301):
            self.assertEqual(f(), 3)
            self.assertEqual(f(1), 4)
            self.assertEqual(f(a=1, b=2), 5)

        # verify maxage
        with mock.patch('dwave.cloud.utils.epochnow', lambda: 299):
            self.assertEqual(f(), 3)
            self.assertEqual(f(1), 4)
            self.assertEqual(f(a=1, b=2), 5)

    def test_default_zero_maxage(self):
        counter = count()

        @self.cached()
        def f(*a, **b):
            return next(counter)

        with mock.patch('dwave.cloud.utils.epochnow', lambda: 0):
            self.assertEqual(f(), 0)
            self.assertEqual(f(), 1)
            self.assertEqual(f(), 2)

    def test_exceptions(self):
        counter = count(0)

        @self.cached()
        def f():
            # raises ZeroDivisionError only on first call
            # we do not want to cache that!
            return 1.0 / next(counter)

        with mock.patch('dwave.cloud.utils.epochnow', lambda: 0):
            self.assertRaises(ZeroDivisionError, f)
            self.assertEqual(f(), 1)
            self.assertEqual(f(), 0.5)


class TestCachedOnDiskDecorator(TestCachedInMemoryDecorator):
    """Test @cached using on-disk store (via @cached.ondisk)."""

    def setUp(self):
        self.tmpdir = tempfile.TemporaryDirectory()
        self.cached = partial(cached.ondisk, directory=self.tmpdir.name)

    def tearDown(self):
        self.tmpdir.cleanup()

    @mock.patch('dwave.cloud.utils.epochnow', lambda: 0)
    def test_persistency(self):
        counter = count()
        def f():
            return next(counter)

        f1 = self.cached(maxage=1)(f)
        self.assertEqual(f1(), 0)

        f2 = self.cached(maxage=1)(f)
        self.assertEqual(f2(), 0)


class TestRetriedDecorator(unittest.TestCase):

    def test_func_called(self):
        """Wrapped function is called with correct arguments."""

        @retried()
        def f(a, b):
            return a, b

        self.assertEqual(f(1, b=2), (1, 2))

    def test_exc_raised(self):
        """Correct exception is raised after number of retries exceeded."""

        @retried()
        def f():
            raise ValueError

        with self.assertRaises(ValueError):
            f()

    def test_func_called_only_until_succeeds(self):
        """Wrapped function is called no more times then it takes to succeed."""

        err = ValueError
        val = mock.sentinel
        attrs = dict(__name__='f')

        # f succeeds on 3rd try
        f = mock.Mock(side_effect=[err, err, val.a, val.b], **attrs)
        ret = retried(3)(f)()
        self.assertEqual(ret, val.a)
        self.assertEqual(f.call_count, 3)

        # fail with only on retry
        f = mock.Mock(side_effect=[err, err, val.a, val.b], **attrs)
        with self.assertRaises(err):
            ret = retried(1)(f)()

        # no errors, return without retries
        f = mock.Mock(side_effect=[val.a, val.b, val.c], **attrs)
        ret = retried(3)(f)()
        self.assertEqual(ret, val.a)
        self.assertEqual(f.call_count, 1)

    def test_decorator(self):
        with self.assertRaises(TypeError):
            retried()("not-a-function")

    def test_backoff_constant(self):
        """Constant retry back-off."""

        # 1s delay before a retry
        backoff = 1

        with mock.patch('time.sleep') as sleep:

            @retried(retries=2, backoff=backoff)
            def f():
                raise ValueError

            with self.assertRaises(ValueError):
                f()

            calls = [mock.call(backoff), mock.call(backoff)]
            sleep.assert_has_calls(calls)

    def test_backoff_seq(self):
        """Retry back-off defined via list."""

        # progressive delay
        backoff = [1, 2, 3]

        with mock.patch('time.sleep') as sleep:

            @retried(retries=3, backoff=backoff)
            def f():
                raise ValueError

            with self.assertRaises(ValueError):
                f()

            calls = [mock.call(b) for b in backoff]
            sleep.assert_has_calls(calls)

    def test_backoff_func(self):
        """Retry back-off defined via callable."""

        def backoff(retry):
            return 2 ** retry

        with mock.patch('time.sleep') as sleep:

            @retried(retries=3, backoff=backoff)
            def f():
                raise ValueError

            with self.assertRaises(ValueError):
                f()

            calls = [mock.call(backoff(1)), mock.call(backoff(2)), mock.call(backoff(3))]
            sleep.assert_has_calls(calls)


class TestDeprecatedDecorator(unittest.TestCase):

    def test_func_called(self):
        """Wrapped function is called with correct arguments."""

        @deprecated()
        def f(a, b):
            return a, b

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")

            self.assertEqual(f(1, b=2), (1, 2))

    def test_warning_raised(self):
        """Correct deprecation message is raised."""

        msg = "deprecation message"

        @deprecated(msg)
        def f():
            return

        with self.assertWarns(DeprecationWarning, msg=msg):
            f()

    def test_warning_raised_automsg(self):
        """Deprecation message is auto-generated and raised."""

        @deprecated()
        def f():
            return

        automsg_regex = r'f\(\) has been deprecated'

        with self.assertWarnsRegex(DeprecationWarning, automsg_regex):
            f()

    def test_decorator(self):
        with self.assertRaises(TypeError):
            deprecated()("not-a-function")


class TestAliasdict(unittest.TestCase):

    def assert_dict_interface(self, aliased, origin):
        "Assert `aliased` behaves exactly as `origin` dict."

        self.assertIsInstance(aliased, dict)
        self.assertDictEqual(aliased, origin)

        self.assertEqual(len(aliased), len(origin))
        self.assertSetEqual(set(aliased), set(origin))
        self.assertSetEqual(set(aliased.keys()), set(origin.keys()))
        self.assertSetEqual(set(aliased.items()), set(origin.items()))
        self.assertListEqual(list(aliased.values()), list(origin.values()))
        for k in origin:
            self.assertIn(k, aliased)
        stranger = "unique-{}".format(''.join(origin))
        self.assertNotIn(stranger, aliased)

        self.assertSetEqual(set(iter(aliased)), set(origin))

        # copy
        new = aliased.copy()
        self.assertIsNot(new, aliased)
        self.assertDictEqual(new, aliased)

        # dict copy constructor on aliasdict
        new = dict(aliased)
        self.assertDictEqual(new, origin)

        # pop on copy
        key = next(iter(origin))
        new.pop(key)
        self.assertSetEqual(set(new), set(origin).difference(key))
        self.assertDictEqual(aliased, origin)

        # get
        self.assertEqual(aliased[key], origin[key])
        self.assertEqual(aliased.get(key), origin.get(key))

        # set
        new = aliased.copy()
        ref = origin.copy()
        new[stranger] = 4
        ref[stranger] = 4
        self.assertDictEqual(new, ref)

        # del
        del new[stranger]
        self.assertDictEqual(new, origin)

    def test_construction(self):
        # aliasdict can be constructed from a mapping/dict
        src = dict(a=1)
        ad = aliasdict(src)
        self.assert_dict_interface(ad, src)

        # aliasdict can be updated, without affecting the source dict
        ad.update(b=2)
        self.assertDictEqual(ad, dict(a=1, b=2))
        self.assertDictEqual(src, dict(a=1))

        # source dict can be updated without affecting the aliased dict
        src.update(c=3)
        self.assertDictEqual(ad, dict(a=1, b=2))
        self.assertDictEqual(src, dict(a=1, c=3))

    def test_dict_interface(self):
        src = dict(a=1, b=2)
        ad = aliasdict(**src)
        self.assert_dict_interface(ad, src)

    def test_alias_concrete(self):
        src = dict(a=1)

        ad = aliasdict(src)
        ad.alias(b=2)

        self.assert_dict_interface(ad, src)
        self.assertEqual(ad['b'], 2)

    def test_alias_callable(self):
        src = dict(a=1)

        ad = aliasdict(src)
        ad.alias(b=lambda d: d.get('a'))

        self.assert_dict_interface(ad, src)

        # 'b' equals 'a'
        self.assertEqual(ad['b'], ad['a'])
        self.assertEqual(ad['b'], src['a'])
        self.assertEqual(ad['b'], 1)

        # it's dynamic, works also when 'a' changes
        ad['a'] = 2
        self.assertEqual(ad['b'], ad['a'])
        self.assertNotEqual(ad['b'], src['a'])
        self.assertEqual(ad['b'], 2)

    def test_alias_get_set_del(self):
        src = dict(a=1)
        ad = aliasdict(src)

        # getitem and get on alias
        ad.alias(b=2)
        self.assertEqual(ad['b'], 2)
        self.assertEqual(ad.get('b'), 2)
        self.assertEqual(ad.aliases['b'], 2)

        # get with a default
        randomkey = str(uuid.uuid4())
        self.assertEqual(ad.get(randomkey), None)
        self.assertEqual(ad.get(randomkey, 1), 1)

        # set alias
        ad['b'] = 3
        self.assertEqual(ad['b'], 3)
        self.assertEqual(ad.aliases['b'], 3)

        # update alias
        ad.alias(b=4)
        self.assertEqual(ad['b'], 4)
        self.assertEqual(ad.aliases['b'], 4)

        # delete alias
        del ad['b']
        self.assertEqual(ad.get('b'), None)

    def test_shadowing(self):
        "Alias keys take precedence."

        ad = aliasdict(a=1)
        ad.alias(a=2)

        self.assertEqual(ad['a'], 2)

        self.assertEqual(dict.__getitem__(ad, 'a'), 1)

    def test_copy(self):
        src = dict(a=1)
        aliases = dict(b=2)

        ad = aliasdict(src)
        ad.alias(**aliases)

        self.assertDictEqual(ad, src)
        self.assertDictEqual(ad.aliases, aliases)

        new = ad.copy()
        self.assertIsInstance(new, aliasdict)
        self.assertIsNot(new, ad)
        self.assertDictEqual(new, src)
        self.assertDictEqual(new.aliases, aliases)

        new = copy.deepcopy(ad)
        self.assertIsInstance(new, aliasdict)
        self.assertIsNot(new, ad)
        self.assertDictEqual(new, src)
        self.assertDictEqual(new.aliases, aliases)


class TestExceptionUtils(unittest.TestCase):

    def raise_implicit():
        try:
            1/0
        except:
            raise ValueError

    def raise_explicit():
        try:
            1/0
        except Exception as e:
            raise ValueError from e

    def raise_mixed():
        try:
            TestExceptionUtils.raise_explicit()
        except:
            raise TypeError

    def test_hasinstance(self):
        # not contained
        self.assertFalse(hasinstance([], ValueError))
        self.assertFalse(hasinstance([TypeError], ValueError))
        self.assertFalse(hasinstance([TypeError()], ValueError))
        self.assertFalse(hasinstance([ValueError], ValueError))

        # contained in unit list
        self.assertTrue(hasinstance([ValueError()], ValueError))
        self.assertTrue(hasinstance([ValueError('msg')], ValueError))

        # contained in a list
        self.assertTrue(hasinstance([TypeError(), ValueError()], ValueError))
        self.assertTrue(hasinstance([ValueError(), ValueError()], ValueError))

        # contained in a tuple
        self.assertTrue(hasinstance((TypeError(), ValueError()), ValueError))

        # base class also contained
        self.assertTrue(hasinstance([ValueError()], Exception))

        # multiple types can be checked
        self.assertFalse(hasinstance([ValueError()], (dict, list)))
        self.assertTrue(hasinstance([ValueError()], (dict, Exception)))
        self.assertTrue(hasinstance([ValueError()], (ValueError, TypeError)))
        self.assertTrue(hasinstance([TypeError()], (ValueError, TypeError)))
        self.assertTrue(hasinstance([TypeError(), ValueError], (ValueError, TypeError)))

    @parameterized.expand([
        (raise_implicit, (ValueError, ZeroDivisionError)),
        (raise_explicit, (ValueError, ZeroDivisionError)),
        (raise_mixed, (TypeError, ValueError, ZeroDivisionError))
    ])
    def test_exception_chain(self, raise_exc, chained_types):
        try:
            raise_exc()
        except Exception as e:
            exc = e
        else:
            self.fail('Exception not raised')

        chain = list(exception_chain(exc))
        self.assertEqual(len(chain), len(chained_types))

        for idx, typ in enumerate(chained_types):
            self.assertIsInstance(chain[idx], typ)

        for typ in chained_types:
            self.assertTrue(hasinstance(exception_chain(exc), typ))

    @parameterized.expand([
        (raise_implicit, (ValueError, ZeroDivisionError)),
        (raise_explicit, (ValueError, ZeroDivisionError)),
        (raise_mixed, (TypeError, ValueError, ZeroDivisionError))
    ])
    def test_is_caused_by(self, raise_exc, exception_types):
        try:
            raise_exc()
        except Exception as exc:
            self.assertTrue(is_caused_by(exc, exception_types))

            for typ in exception_types:
                self.assertTrue(is_caused_by(exc, typ))

            self.assertFalse(is_caused_by(exc, KeyError))


class TestFilteredSecretsFormatter(unittest.TestCase):

    def _filter(self, msg):
        fmt = FilteredSecretsFormatter('%(msg)s')
        rec = logging.makeLogRecord(dict(msg=msg))
        return fmt.format(rec)

    @parameterized.expand([
        ('0123456789012345678901234567890123456789', None),
        ('A-0123456789012345678901234567890123456789', None),
        ('AB-0123456789012345678901234567890123456789', 'AB-012...789'),
        ('ABC-0123456789012345678901234567890123456789', 'ABC-012...789'),
        ('ABC-c0f3456789012345678901234567890123456fee', 'ABC-c0f...fee'),
        ('ABC-c0f3456789012345678901234567890123456beeffee', 'ABC-c0f...fee'),
        ('ABC-c0f3456789012345678901234567890123456beefxfee', None),
        ('ABCD-0123456789012345678901234567890123456789', 'ABCD-012...789'),
        ('ABCDE-0123456789012345678901234567890123456789', None),
    ])
    def test_sapi_tokens(self, inp, out=None):
        if out is None:
            out = inp

        self.assertEqual(self._filter(inp), out)

        ctx = '{} suffix'
        self.assertEqual(self._filter(ctx.format(inp)), ctx.format(out))

        ctx = 'prefix {}'
        self.assertEqual(self._filter(ctx.format(inp)), ctx.format(out))

        ctx = 'a{}word'
        self.assertEqual(self._filter(ctx.format(inp)), ctx.format(inp))


if __name__ == '__main__':
    unittest.main()



# Copyright 2021 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import enum


# Default API endpoints
DEFAULT_SOLVER_API_ENDPOINT = 'https://cloud.dwavesys.com/sapi/'

DEFAULT_METADATA_API_ENDPOINT = 'https://cloud.dwavesys.com/metadata/v1/'

DEFAULT_REGION = 'na-west-1'


class ProblemStatus(str, enum.Enum):
    """Solver API problem status values.

    Initially a problem is in the PENDING state. When the D-Wave system starts
    to process a problem, its state changes to IN_PROGRESS. After completion,
    the problem status changes to either COMPLETED or FAILED (if an error
    occurred). COMPLETED, FAILED, and CANCELLED are all terminal states.

    After a problem enters a terminal state, its status does not change. Users
    can cancel a problem at any time before it reaches its terminal state.
    """

    PENDING = "PENDING"
    IN_PROGRESS = "IN_PROGRESS"
    COMPLETED = "COMPLETED"
    FAILED = "FAILED"
    CANCELLED = "CANCELLED"


class ProblemEncodingFormat(str, enum.Enum):
    QP = "qp"
    BQ = "bq"   # deprecated for submission
    REF = "ref"


class AnswerEncodingFormat(str, enum.Enum):
    QP = "qp"
    BQ = "bq"   # dimod (de-)serialization-based


class ProblemType(str, enum.Enum):
    ISING = "ising"
    QUBO = "qubo"
    BQM = "bqm"
    CQM = "cqm"
    DQM = "dqm"



# Copyright 2021 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# NOTE: dwave.cloud.api module and submodules considered private for now!


import dwave.cloud.api.client
import dwave.cloud.api.constants
import dwave.cloud.api.exceptions
import dwave.cloud.api.models
import dwave.cloud.api.resources

from dwave.cloud.api import client
from dwave.cloud.api import constants
from dwave.cloud.api import exceptions
from dwave.cloud.api import models
from dwave.cloud.api import resources

from dwave.cloud.api.client import DWaveAPIClient, SolverAPIClient, MetadataAPIClient
from dwave.cloud.api.resources import Solvers, Problems, Regions



# Copyright 2021 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import json
import uuid
import unittest
from urllib.parse import urljoin, urlparse, parse_qs
from itertools import chain

import numpy
import requests_mock

from parameterized import parameterized

try:
    import dimod
except ImportError:
    dimod = None

from dwave.cloud.client.base import Client
from dwave.cloud.coders import (
    encode_problem_as_qp, decode_qp,
    encode_problem_as_ref, decode_bq)
from dwave.cloud.api.resources import Problems
from dwave.cloud.api import exceptions, models, constants

from tests import config
from tests.api.mocks import StructuredSapiMockResponses, UnstructuredSapiMockResponses


class AssertionSatisfied(Exception):
    """Raise it to quickly unwrap the stack and signal assertion satisfied."""


class ProblemResourcesBaseTests(abc.ABC):
    """Basic tests for `dwave.cloud.api.resources.Problems`."""

    @classmethod
    def setUpClass(cls):
        """Create and submit one problem.

        Set the following attributes:
        - `linear`/`quadratic` (or `bqm`), `problem_data`
        - `problem_type`, `problem_id`, `problem_label`
        - `params`
        - `solver_id`

        Written to work with:
        - both mock data and live api
        - both structured and unstructured problems/answers/solvers
        """

    def verify_problem_status(self, status: models.ProblemStatus, solved: bool = False):
        """Verify `status` consistent with the submitted problem."""
        self.assertEqual(status.id, self.problem_id)
        self.assertEqual(status.type, self.problem_type)
        self.assertEqual(status.solver, self.solver_id)
        self.assertEqual(status.label, self.problem_label)
        self.assertEqual(status.status, constants.ProblemStatus.COMPLETED)
        self.assertIsNotNone(status.submitted_on)
        if solved:
            self.assertIsNotNone(status.solved_on)

    @abc.abstractmethod
    def verify_problem_answer(self, answer: models.ProblemAnswer):
        """Verify `answer` consistent with the submitted problem."""

    def test_list_all(self):
        """List of all available problems retrieved."""

        ps = self.api.list_problems()
        self.assertIsInstance(ps, list)
        self.assertGreater(len(ps), 0)
        self.assertLessEqual(len(ps), 1000)     # sapi limit

        p = ps.pop()
        self.assertIsInstance(p, models.ProblemStatus)

    def test_list_filter(self):
        """Problem filtering works (on edges as well)."""

        # 404 on non-existing problem id
        with self.assertRaises(exceptions.ResourceNotFoundError):
            self.api.list_problems(id='nonexisting-problem-id')

        # result size limited
        ps = self.api.list_problems(max_results=1)
        self.assertEqual(len(ps), 1)

        # 400 on filtering by invalid status
        with self.assertRaises(exceptions.ResourceBadRequestError):
            self.api.list_problems(status='nonexisting-status-id')

        # empty result for non-existing solver
        ps = self.api.list_problems(solver='nonexisting-solver-id')
        self.assertEqual(len(ps), 0)

        # filtering by valid status
        ps = self.api.list_problems(status=constants.ProblemStatus.COMPLETED)
        self.assertGreater(len(ps), 0)

        # extract one completed problem for further tests
        problem = ps.pop()

        # filter by exact id
        ps = self.api.list_problems(id=problem.id)
        self.assertEqual(len(ps), 1)

        # verify all fields of the returned problem are identical to the original
        p = ps.pop()
        self.assertEqual(p, problem)

    def test_get_problem(self):
        """Problem status with answer retrieved by problem id."""

        problem_id = self.problem_id
        status = self.api.get_problem(problem_id)

        self.assertIsInstance(status, models.ProblemStatusMaybeWithAnswer)
        self.verify_problem_status(status, solved=True)

        self.assertIsInstance(status.answer, models.ProblemAnswer)
        self.verify_problem_answer(status.answer)

    def test_get_problem_status(self):
        """Problem status retrieved by problem id."""

        problem_id = self.problem_id
        status = self.api.get_problem_status(problem_id)

        self.assertIsInstance(status, models.ProblemStatus)
        self.verify_problem_status(status, solved=True)

    def test_get_problem_statuses(self):
        """Multiple problem statuses retrieved by problem ids."""

        problem_id = self.problem_id
        ids = [problem_id] * 3
        statuses = self.api.get_problem_statuses(ids)

        self.assertEqual(len(statuses), len(ids))
        for status in statuses:
            self.assertIsInstance(status, models.ProblemStatus)
            self.verify_problem_status(status, solved=True)

    def test_get_problem_info(self):
        """Problem info (complete problem description) retrieved by problem id."""

        problem_id = self.problem_id
        info = self.api.get_problem_info(problem_id)

        self.assertIsInstance(info, models.ProblemInfo)
        self.assertEqual(info.id, problem_id)

        # data
        self.assertIsInstance(info.data, models.ProblemData)
        self.assertEqual(info.data, self.problem_data)

        # params
        self.assertEqual(info.params, self.params)

        # metadata
        self.assertIsInstance(info.metadata, models.ProblemMetadata)
        self.assertEqual(info.metadata.solver, self.solver_id)
        self.assertEqual(info.metadata.type, self.problem_type)
        self.assertEqual(info.metadata.label, self.problem_label)
        self.assertEqual(info.metadata.status, constants.ProblemStatus.COMPLETED)
        self.assertEqual(info.metadata.submitted_by, self.token)
        self.assertIsNotNone(info.metadata.submitted_on)
        self.assertIsNotNone(info.metadata.solved_on)

        # answer
        self.assertIsInstance(info.answer, models.ProblemAnswer)
        self.verify_problem_answer(info.answer)

    def test_get_problem_answer(self):
        """Problem answer retrieved by problem id."""

        problem_id = self.problem_id
        answer = self.api.get_problem_answer(problem_id)

        self.assertIsInstance(answer, models.ProblemAnswer)
        self.verify_problem_answer(answer)

    def test_get_problem_messages(self):
        """Problem messages retrieved by problem id."""

        problem_id = self.problem_id
        messages = self.api.get_problem_messages(problem_id)

        self.assertIsInstance(messages, list)
        if messages:
            self.assertIsInstance(messages[0], dict)

    def test_problem_submit(self):
        """Problem submitted."""

        status = self.api.submit_problem(
            data=self.problem_data,
            params=self.params,
            solver=self.solver_id,
            type=self.problem_type,
        )

        self.assertIsInstance(status, models.ProblemStatusMaybeWithAnswer)
        self.assertEqual(status.type, self.problem_type)
        self.assertEqual(status.solver, self.solver_id)
        self.assertIsNotNone(status.submitted_on)

        if status.status is constants.ProblemStatus.COMPLETED:
            self.assertIsInstance(status.answer, models.ProblemAnswer)
            self.verify_problem_answer(status.answer)

    def test_problem_submit_errors(self):
        """Problem submit fails due to invalid parameters."""

        with self.assertRaises(exceptions.ResourceBadRequestError):
            self.api.submit_problem(
                data=self.problem_data,
                params=dict(non_existing_param=1),
                solver=self.solver_id,
                type=self.problem_type,
            )

        with self.assertRaises(exceptions.ResourceNotFoundError):
            self.api.submit_problem(
                data=self.problem_data,
                params=self.params,
                solver='non-existing-solver',
                type=self.problem_type,
            )

    def test_problem_batch_submit(self):
        """Multiple problems are submitted and initial statuses returned."""

        job = models.ProblemJob(
            data=self.problem_data,
            params=self.params,
            solver=self.solver_id,
            type=self.problem_type,
        )

        statuses = self.api.submit_problems([job] * 3)

        self.assertIsInstance(statuses, list)
        self.assertEqual(len(statuses), 3)

        for status in statuses:
            self.assertIsInstance(status, models.ProblemInitialStatus)

    def test_problem_batch_submit_error(self):
        """Problem batch submit fails due to invalid parameters."""

        job = models.ProblemJob(
            data=self.problem_data,
            params=dict(non_existing_param=1),
            solver=self.solver_id,
            type=self.problem_type,
        )

        statuses = self.api.submit_problems([job])

        self.assertIsInstance(statuses, list)
        self.assertEqual(len(statuses), 1)

        for status in statuses:
            self.assertIsInstance(status, models.ProblemSubmitError)
            self.assertEqual(status.error_code, 400)

    def test_problem_cancel(self):
        """(Finished) problem cancel attempted."""

        with self.assertRaises(exceptions.ResourceConflictError):
            self.api.cancel_problem(self.problem_id)

    def test_problem_batch_cancel(self):
        """(Finished) problem cancel attempted."""

        problem_ids = [self.problem_id, self.problem_id]
        statuses = self.api.cancel_problems(problem_ids)

        self.assertIsInstance(statuses, list)
        self.assertEqual(len(statuses), len(problem_ids))

        for status in statuses:
            self.assertIsInstance(status, models.ProblemCancelError)
            self.assertEqual(status.error_code, 409)


class StructuredProblemTestsMixin:

    def verify_problem_answer(self, answer: models.ProblemAnswer):
        ans = decode_qp(msg=dict(answer=answer.dict(), type=self.problem_type.value))
        var = set(chain(*self.quadratic)) | self.linear.keys()
        self.assertEqual(set(ans['active_variables']), var)
        self.assertEqual(len(ans['energies']), len(ans['solutions']))
        self.assertEqual(sum(ans['num_occurrences']), self.params['num_reads'])


class UnstructuredProblemTestsMixin:

    def verify_problem_answer(self, answer: models.ProblemAnswer):
        ans = decode_bq(msg=dict(answer=answer.dict(), type=self.problem_type.value))
        ss = ans['sampleset']
        self.assertEqual(ss.variables, self.bqm.variables)
        dimod.testing.assert_sampleset_energies(ss, self.bqm)


class ProblemResourcesMockerMixin:
    """Define and set up mock responses for all requests issued by
    `ProblemResourcesBaseTests` tests.
    """

    token = str(uuid.uuid4())
    endpoint = 'http://test.com/path/'

    def setUp(self):
        def url(path):
            return urljoin(self.endpoint, path)

        self.mocker = requests_mock.Mocker()

        headers = {'X-Auth-Token': self.token}
        self.mocker.get(requests_mock.ANY, status_code=401)
        self.mocker.get(requests_mock.ANY, status_code=404, request_headers=headers)

        # p1 is completed
        p1_id = self.p1.problem_id
        p1_status = self.p1.complete_no_answer_reply()
        p1_status_with_answer = self.p1.complete_reply()
        p1_params = self.params
        p1_metadata = self.p1.problem_metadata(submitted_by=self.token)
        p1_answer = self.p1.problem_answer()
        p1_messages = self.p1.problem_messages()
        p1_info = self.p1.problem_info(params=p1_params, metadata=p1_metadata)

        # p2 is submitted (and pending)
        p2_id = self.p2.problem_id
        p2_status = self.p2.continue_reply(id=p2_id)

        # p3 is cancelled
        p3_id = self.p3.problem_id
        p3_status = self.p3.cancel_reply(id=p3_id)

        all_problem_ids = {p1_id, p2_id, p3_id}
        all_solvers = [self.solver_id]
        all_statuses = {s.value for s in constants.ProblemStatus}

        # problem status et al
        self.mocker.get(
            url(f'problems/{p1_id}'),
            json=p1_status_with_answer,
            request_headers=headers)
        self.mocker.get(
            url(f'problems/?id={p1_id}'),
            complete_qs=True,
            json=[p1_status],
            request_headers=headers)
        self.mocker.get(
            url(f'problems/{p1_id}/info'),
            json=p1_info,
            request_headers=headers)
        self.mocker.get(
            url(f'problems/{p1_id}/answer'),
            json=p1_answer,
            request_headers=headers)
        self.mocker.get(
            url(f'problems/{p1_id}/messages'),
            json=p1_messages,
            request_headers=headers)
        self.mocker.get(
            url(f'problems/?id={p1_id},{p1_id},{p1_id}'),
            complete_qs=True,
            json=[p1_status] * 3,
            request_headers=headers)

        # list problems
        self.mocker.get(
            url('problems/'),
            complete_qs=True,
            json=[p1_status, p2_status, p3_status],
            request_headers=headers)
        self.mocker.get(
            url(f'problems/?id={p1_id}'),
            complete_qs=True,
            json=[p1_status],
            request_headers=headers)
        self.mocker.get(
            url(f'problems/?id={p1_id},{p2_id}'),
            complete_qs=True,
            json=[p1_status, p2_status],
            request_headers=headers)
        self.mocker.get(
            url('problems/?max_results=1'),
            json=[p1_status],
            request_headers=headers)
        self.mocker.get(
            url('problems/?status=COMPLETED'),
            json=[p1_status],
            request_headers=headers)

        def match_invalid_status(request):
            query = parse_qs(urlparse(request.url).query)
            status = query.get('status')
            return status is not None and status[0] not in all_statuses

        self.mocker.get(
            url('problems/'),
            additional_matcher=match_invalid_status,
            status_code=400,
            request_headers=headers)

        def match_invalid_solver(request):
            query = parse_qs(urlparse(request.url).query)
            solver = query.get('solver')
            return solver is not None and solver[0] not in all_solvers

        self.mocker.get(
            url('problems/'),
            additional_matcher=match_invalid_solver,
            json=[],
            request_headers=headers)

        def match_invalid_problem_id(request):
            query = parse_qs(urlparse(request.url).query)
            problem_id = query.get('id')
            return problem_id is not None and problem_id[0] not in all_problem_ids

        self.mocker.get(
            url('problems/'),
            complete_qs=True,
            additional_matcher=match_invalid_problem_id,
            status_code=404,
            request_headers=headers)

        # problem submit
        self.mocker.post(url('problems/'), json=p2_status, request_headers=headers)

        def match_invalid_problem_params(request):
            return request.json()['params'] != self.params

        self.mocker.post(
            url('problems/'),
            additional_matcher=match_invalid_problem_params,
            status_code=400,
            request_headers=headers)

        def match_invalid_problem_solver(request):
            return request.json()['solver'] != self.solver_id

        self.mocker.post(
            url('problems/'),
            additional_matcher=match_invalid_problem_solver,
            status_code=404,
            request_headers=headers)

        # problem batch submit
        def match_batch_submit(request):
            data = request.json()
            return isinstance(data, list) and len(data) == 3

        self.mocker.post(
            url('problems/'),
            additional_matcher=match_batch_submit,
            json=[p2_status] * 3,
            request_headers=headers)

        def match_invalid_batch_submit(request):
            data = request.json()
            return isinstance(data, list) and len(data) == 1 and data[0]['params'] != self.params

        self.mocker.post(
            url('problems/'),
            additional_matcher=match_invalid_batch_submit,
            json=[self.p1.immediate_error_reply(code=400, msg='Unknown parameter')],
            request_headers=headers)

        # problem cancel
        self.mocker.delete(url(f'problems/{p1_id}'), status_code=409, request_headers=headers)
        self.mocker.delete(
            url('problems/'),
            json=[self.p1.immediate_error_reply(code=409, msg='Problem has been finished.')] * 2,
            request_headers=headers)

        self.mocker.start()

        self.api = Problems(token=self.token, endpoint=self.endpoint)

    def tearDown(self):
        self.mocker.stop()

        self.api.close()


class TestMockProblemsStructured(StructuredProblemTestsMixin,
                                 ProblemResourcesMockerMixin,
                                 ProblemResourcesBaseTests,
                                 unittest.TestCase):
    """Verify `dwave.cloud.api.resources.Problems` handle structured problems
    using mocked SAPI responses.
    """

    @classmethod
    def setUpClass(cls):
        """Configure attributes required (used) by ProblemResourcesBaseTests."""

        cls.p1 = StructuredSapiMockResponses()
        cls.p2 = StructuredSapiMockResponses()
        cls.p3 = StructuredSapiMockResponses()

        cls.linear, cls.quadratic = cls.p1.problem
        cls.problem_data = models.ProblemData.parse_obj(cls.p1.problem_data())
        cls.problem_type = constants.ProblemType(cls.p1.problem_type)
        cls.problem_id = cls.p1.problem_id
        cls.problem_label = cls.p1.problem_label

        cls.params = dict(num_reads=100)
        cls.solver_id = cls.p1.solver.id


@unittest.skipUnless(dimod, "dimod not installed")
class TestMockProblemsUnstructured(UnstructuredProblemTestsMixin,
                                   ProblemResourcesMockerMixin,
                                   ProblemResourcesBaseTests,
                                   unittest.TestCase):
    """Verify `dwave.cloud.api.resources.Problems` handle unstructured problems
    using mocked SAPI responses.
    """

    @classmethod
    def setUpClass(cls):
        """Configure attributes required (used) by ProblemResourcesBaseTests."""

        cls.p1 = UnstructuredSapiMockResponses()
        cls.p2 = UnstructuredSapiMockResponses()
        cls.p3 = UnstructuredSapiMockResponses()

        cls.bqm = cls.p1.problem
        cls.problem_data = models.ProblemData.parse_obj(cls.p1.problem_data())
        cls.problem_type = constants.ProblemType(cls.p1.problem_type)
        cls.problem_id = cls.p1.problem_id
        cls.problem_label = cls.p1.problem_label

        cls.params = dict(time_limit=3)
        cls.solver_id = cls.p1.solver.id


@unittest.skipUnless(config, "SAPI access not configured")
class TestCloudProblemsStructured(StructuredProblemTestsMixin,
                                  ProblemResourcesBaseTests,
                                  unittest.TestCase):
    """Verify `dwave.cloud.api.resources.Problems` handle structured problems
    as returned by SAPI.
    """

    @classmethod
    def setUpClass(cls):
        """Configure attributes required (used) by ProblemResourcesBaseTests."""

        with Client(**config) as client:
            cls.token = config['token']
            cls.api = Problems.from_client_config(client)

            # submit and solve an Ising problem as a fixture
            solver = client.get_solver(qpu=True)
            cls.solver_id = solver.id
            edge = next(iter(solver.edges))
            cls.linear = {}
            cls.quadratic = {edge: 1.0}
            qp = encode_problem_as_qp(solver, cls.linear, cls.quadratic)
            cls.problem_data = models.ProblemData.parse_obj(qp)
            cls.problem_type = constants.ProblemType.ISING
            cls.params = dict(num_reads=100)
            future = solver.sample_ising(cls.linear, cls.quadratic, **cls.params)
            resolved = future.result()
            cls.problem_id = future.id
            cls.problem_label = future.label

            # double-check
            assert future.remote_status == constants.ProblemStatus.COMPLETED.value

    @classmethod
    def tearDownClass(cls):
        cls.api.close()


@unittest.skipUnless(dimod, "dimod not installed")
@unittest.skipUnless(config, "SAPI access not configured")
class TestCloudProblemsUnstructured(UnstructuredProblemTestsMixin,
                                    ProblemResourcesBaseTests,
                                    unittest.TestCase):
    """Verify `dwave.cloud.api.resources.Problems` handle unstructured problems
    as returned by SAPI.
    """

    @classmethod
    def setUpClass(cls):
        """Configure attributes required (used) by ProblemResourcesBaseTests."""

        with Client(**config) as client:
            cls.token = config['token']
            cls.api = Problems.from_client_config(client)

            # submit and solve an Ising problem as a fixture
            solver = client.get_solver(hybrid=True)
            cls.solver_id = solver.id
            cls.bqm = dimod.BQM.from_ising({}, {'ab': 1.0})
            problem_data_id = solver.upload_problem(cls.bqm).result()
            problem_data_ref = encode_problem_as_ref(problem_data_id)
            cls.problem_data = models.ProblemData.parse_obj(problem_data_ref)
            cls.problem_type = constants.ProblemType.BQM
            cls.params = dict(time_limit=3)
            future = solver.sample_bqm(problem_data_id, **cls.params)
            resolved = future.result()
            cls.problem_id = future.id
            cls.problem_label = future.label

            # double-check
            assert future.remote_status == constants.ProblemStatus.COMPLETED.value

    @classmethod
    def tearDownClass(cls):
        cls.api.close()


class NumpyParamsSerialization(unittest.TestCase):

    # basic scalar types we support for parameter values
    NUMPY_TYPES_AS_PYTHON = [
        (numpy.bool_(1), True), (numpy.bool8(1), True),
        (numpy.byte(1), 1), (numpy.int8(1), 1),
        (numpy.ubyte(1), 1), (numpy.uint8(1), 1),
        (numpy.short(1), 1), (numpy.int16(1), 1),
        (numpy.ushort(1), 1), (numpy.uint16(1), 1),
        (numpy.intc(1), 1), (numpy.int32(1), 1),
        (numpy.uintc(1), 1), (numpy.uint32(1), 1),
        (numpy.int_(1), 1), (numpy.int32(1), 1),
        (numpy.uint(1), 1), (numpy.uint32(1), 1),
        (numpy.longlong(1), 1), (numpy.int64(1), 1),
        (numpy.ulonglong(1), 1), (numpy.uint64(1), 1),
        (numpy.half(1.0), 1.0), (numpy.float16(1.0), 1.0),
        (numpy.single(1.0), 1.0), (numpy.float32(1.0), 1.0),
        (numpy.double(1.0), 1.0), (numpy.float64(1.0), 1.0),
        (numpy.longdouble(1.0), 1.0)
    ] + ([
        (numpy.float128(1.0), 1.0)      # unavailable on windows
    ] if hasattr(numpy, 'float128') else [
    ])

    @classmethod
    def setUpClass(cls):
        _p = StructuredSapiMockResponses()
        cls.linear, cls.quadratic = _p.problem
        cls.problem_data = models.ProblemData.parse_obj(_p.problem_data())
        cls.problem_type = constants.ProblemType(_p.problem_type)
        cls.problem_id = _p.problem_id
        cls.problem_label = _p.problem_label
        cls.params = {}
        cls.solver_id = _p.solver.id
        cls.api = Problems(token='token', endpoint='http://end.point/')

    @parameterized.expand(NUMPY_TYPES_AS_PYTHON)
    def test_problem_submit(self, np_val, py_val):
        user_params = dict(num_reads=np_val)
        expected_params = dict(num_reads=py_val)

        def verify_params(*args, **kwargs):
            body = json.loads(kwargs.get('data'))
            params = body.get('params')
            if params != expected_params:
                raise AssertionError("params don't match")
            raise AssertionSatisfied

        with unittest.mock.patch.object(self.api.session, 'post', verify_params):
            with self.assertRaises(AssertionSatisfied):
                self.api.submit_problem(
                    data=self.problem_data,
                    params=user_params,
                    solver=self.solver_id,
                    type=self.problem_type,
                )

    @parameterized.expand(NUMPY_TYPES_AS_PYTHON)
    def test_problem_batch_submit(self, np_val, py_val):
        user_params = dict(num_reads=np_val)
        expected_params = dict(num_reads=py_val)

        def verify_params(*args, **kwargs):
            body = json.loads(kwargs.get('data'))
            params = body[0].get('params')
            if params != expected_params:
                raise AssertionError("params don't match")
            raise AssertionSatisfied

        job = models.ProblemJob(
            data=self.problem_data,
            params=user_params,
            solver=self.solver_id,
            type=self.problem_type,
        )

        with unittest.mock.patch.object(self.api.session, 'post', verify_params):
            with self.assertRaises(AssertionSatisfied):
                self.api.submit_problems([job])



# Copyright 2017 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Computation manages the interactions between your code and a :term:`solver`, which
manages interactions between the remote resource and your submitted problems.

Your solver instantiates a :class:`Future` object for its calls, via D-Wave Sampler API
(SAPI) servers, to the remote resource.

You can interact through the :class:`Future` object with pending (running) or completed
computation---sampling on a QPU or software solver---executed remotely, monitoring problem status,
waiting for and retrieving results, cancelling enqueued jobs, etc.

Some :class:`Future` methods are blocking.

"""

import time
import threading
import functools

from operator import itemgetter
from dateutil.parser import parse
from concurrent.futures import TimeoutError

from dwave.cloud.utils import (
    utcnow, datetime_to_timestamp, aliasdict, deprecated)
from dwave.cloud.exceptions import InvalidAPIResponseError

# Use numpy if available for fast decoding
try:
    import numpy as np
    _numpy = True
except ImportError:
    _numpy = False

__all__ = ['Future']


@functools.total_ordering
class Future(object):
    """Class for interacting with jobs submitted to SAPI.

    :class:`~dwave.cloud.solver.Solver` uses :class:`Future` to construct
    objects for pending SAPI calls that can wait for requests to complete and
    parse returned messages.

    Objects are blocked for the duration of any data accessed on the remote
    resource.

    Warning:
        :class:`Future` objects are not intended to be directly
        created. Problem submittal is initiated by one of the solvers in
        :mod:`~dwave.cloud.solver` module and executed by one of the clients.

    Args:
        solver (:class:`~dwave.cloud.solver.Solver`):
            Solver responsible for this :class:`Future` object.

        id_ (str, optional, default=None):
            Identification for a query submitted by a solver to SAPI. May be
            None following submission until an identification number is set.

        return_matrix (bool, optional, default=False):
            Return values for this :class:`Future` object are NumPy matrices.

    Examples:
        This example creates a solver using the local system's default D-Wave
        Cloud Client configuration file, submits a simple QUBO problem to a
        remote D-Wave resource for 100 samples, and checks a couple of times
        whether the sampling is completed.

        >>> from dwave.cloud import Client
        >>> client = Client.from_config()       # doctest: +SKIP
        >>> solver = client.get_solver()        # doctest: +SKIP
        >>> u, v = next(iter(solver.edges))     # doctest: +SKIP
        >>> Q = {(u, u): -1, (u, v): 0, (v, u): 2, (v, v): -1}   # doctest: +SKIP
        >>> computation = solver.sample_qubo(Q, num_reads=100)   # doctest: +SKIP
        >>> computation.done()  # doctest: +SKIP
        False
        >>> computation.id   # doctest: +SKIP
        '1cefeb6d-ebd5-4592-87c0-4cc43ec03e27'
        >>> computation.done()   # doctest: +SKIP
        True
        >>> client.close()       # doctest: +SKIP
    """

    def __init__(self, solver, id_, return_matrix=False):
        self.solver = solver

        # Has the client tried to cancel this job
        self._cancel_requested = False
        self._cancel_sent = False
        self._single_cancel_lock = threading.Lock()  # Make sure we only call cancel once

        # ID readiness notification
        self._id_ready_event = threading.Event()

        # Should the results be decoded as python lists or numpy matrices
        if return_matrix and not _numpy:
            raise ValueError("Matrix result requested without numpy.")
        self.return_matrix = return_matrix

        #: The id the server will use to identify this problem, None until the id is actually known
        self.id = id_

        #: Problem label, as (optionally) set on submission. None until parsed from a response.
        self.label = None

        #: `datetime` the Future was created (immediately before enqueued in Client's submit queue)
        self.time_created = utcnow()

        #: `datetime` corresponding to the time when the problem was accepted by the server (None before then)
        self.time_received = None

        #: `datetime` corresponding to the time when the problem was completed by the server (None before then)
        self.time_solved = None

        #: `datetime` the Future was resolved (marked as done; succeeded or failed), or None before then
        self.time_resolved = None

        # Track how long it took us to parse the data
        self.parse_time = None

        # approx. server-client clocks difference in seconds
        self.clock_diff = None

        # Data from the server before it is parsed
        self._message = None

        #: Status flag most recently returned by the server
        self.remote_status = None

        # Data from the server after it's parsed
        self._result = None
        self._exception = None

        # Event(s) to signal when the results are ready
        self._results_ready_event = threading.Event()
        self._other_events = []

        # current poll back-off interval, in seconds
        self._poll_backoff = None

        # XXX: energy offset carried via Future, until implemented in SAPI
        self._offset = 0

    # make Future ordered

    def __lt__(self, other):
        return id(self) < id(other)

    def __eq__(self, other):
        return self is other

    def __hash__(self):
        return id(self)

    def _set_message(self, message):
        """Complete the future with a message from the server.

        The message from the server may actually be an error.

        Args:
            message (dict):
                Data from the server from trying to complete query.
        """
        self._message = message
        self._signal_ready()

    def _set_exception(self, exception):
        """Complete the future with an exception.

        Args:
            exception (Exception):
                Exception that caused the failure.
        """
        self._exception = exception
        self._signal_ready()

    def _signal_ready(self):
        """Signal all the events waiting on this future."""
        self.time_resolved = utcnow()
        self._id_ready_event.set()
        self._results_ready_event.set()
        [ev.set() for ev in self._other_events]

    def _add_event(self, event):
        """Add an event to be signaled after this event completes."""
        self._other_events.append(event)
        if self.done():
            event.set()

    def _remove_event(self, event):
        """Remove a completion event from this future."""
        self._other_events.remove(event)

    def _set_clock_diff(self, server_response, localtime_of_response):
        """Calculate and set the `.clock_diff`, based on headers from a server
        response, and the local time of response received.
        """
        try:
            server_time = datetime_to_timestamp(parse(server_response.headers['date']))
        except:
            server_time = 0
        self.clock_diff = abs(server_time - localtime_of_response)

    @staticmethod
    def wait_multiple(futures, min_done=None, timeout=None):
        """Wait for multiple :class:`Future` objects to complete.

        Blocking call that uses an event object to emulate multi-wait for Python.

        Args:
            futures (list of Futures):
                List of :class:`Future` objects to await.

            min_done (int, optional, default=None):
                Minimum required completions to end the waiting. The wait is
                terminated when this number of results are ready. If None, waits
                for all the :class:`Future` objects to complete.

            timeout (float, optional, default=None):
                Maximum number of seconds to await completion. If None, waits
                indefinitely.

        Returns:
            Two-tuple of :class:`Future` objects: completed and not completed
            submitted tasks. Similar to `concurrent.futures.wait()` method's
            returned two-tuple of `done` and `not_done` sets.

        See Also:
            :func:`as_completed` for a blocking iterable of resolved futures
            similar to `concurrent.futures.as_completed()` method.

        Examples:
            This example creates a solver using the local system's default
            D-Wave Cloud Client configuration file, submits a simple QUBO
            problem to a remote D-Wave resource 3 times for differing numers of
            samples, and waits for sampling to complete on any two of the
            submissions. The wait ends with the completion of two submissions
            while the third is still in progress. (A more typical approach would
            use something like
            :code:`first = next(Future.as_completed(computation))` instead.)

            >>> import dwave.cloud as dc
            >>> client = dc.Client.from_config()  # doctest: +SKIP
            >>> solver = client.get_solver()      # doctest: +SKIP
            >>> u, v = next(iter(solver.edges))   # doctest: +SKIP
            >>> Q = {(u, u): -1, (u, v): 0, (v, u): 2, (v, v): -1}    # doctest: +SKIP
            >>> computation = [solver.sample_qubo(Q, num_reads=1000),
            ...                solver.sample_qubo(Q, num_reads=50),
            ...                solver.sample_qubo(Q, num_reads=10)]   # doctest: +SKIP
            >>> dc.computation.Future.wait_multiple(computation, min_done=1)    # doctest: +SKIP
            ([<dwave.cloud.computation.Future at 0x17dde518>,
              <dwave.cloud.computation.Future at 0x17ddee80>],
             [<dwave.cloud.computation.Future at 0x15078080>])
            >>> print(computation[0].done())   # doctest: +SKIP
            False
            >>> print(computation[1].done())  # doctest: +SKIP
            True
            >>> print(computation[2].done())   # doctest: +SKIP
            True
            >>> client.close()         # doctest: +SKIP

        """
        if min_done is None:
            min_done = len(futures)

        if timeout is None:
            timeout = float('inf')

        # Track the exit conditions
        finish = time.time() + timeout
        done = []

        # Keep track of what futures haven't finished
        remaining = list(futures)

        # Insert our event into all the futures
        event = threading.Event()
        [f._add_event(event) for f in remaining]

        # Check the exit conditions
        while len(done) < min_done and finish > time.time():
            # Prepare to wait on any of the jobs finishing
            event.clear()

            # Check if any of the jobs have finished. After the clear just in
            # case one finished and we erased the signal it by calling clear above
            finished_futures = {f for f in remaining if f.done()}
            if len(finished_futures) > 0:
                # If we did make a mistake reseting the event, undo that now
                # so that we double check the finished list before a wait blocks
                event.set()

                # Update our exit conditions
                done.extend(finished_futures)
                remaining = [f for f in remaining if f not in finished_futures]
                continue

            # Block on any of the jobs finishing
            wait_time = finish - time.time() if abs(finish) != float('inf') else None
            event.wait(wait_time)

        # Clean up after ourselves
        [f._remove_event(event) for f in futures]
        return done, remaining

    @staticmethod
    def as_completed(fs, timeout=None):
        """Yield Futures objects as they complete.

        Returns an iterator over the specified list of :class:`Future` objects
        that yields those objects as they complete. Completion occurs when the
        submitted job is finished or cancelled.

        Emulates the behavior of the `concurrent.futures.as_completed()`
        function.

        Args:
            fs (list):
                List of :class:`Future` objects to iterate over.

            timeout (float, optional, default=None):
                Maximum number of seconds to await completion. If None, awaits
                indefinitely.

        Returns:
            Generator (:class:`Future` objects):
                Listed :class:`Future` objects as they complete.

        Raises:
            `concurrent.futures.TimeoutError` is raised if per-future timeout is
            exceeded.

        Examples:
            This example creates a solver using the local system's default D-Wave
            Cloud Client configuration file, submits a simple QUBO problem to a
            remote D-Wave resource 3 times for differing numers of samples, and
            yields timing information for each job as it completes.

            >>> import dwave.cloud as dc
            >>> client = dc.Client.from_config()   # doctest: +SKIP
            >>> solver = client.get_solver()       # doctest: +SKIP
            >>> u, v = next(iter(solver.edges))    # doctest: +SKIP
            >>> Q = {(u, u): -1, (u, v): 0, (v, u): 2, (v, v): -1}    # doctest: +SKIP
            >>> computation = [solver.sample_qubo(Q, num_reads=1000),
            ...                solver.sample_qubo(Q, num_reads=50),
            ...                solver.sample_qubo(Q, num_reads=10)]   # doctest: +SKIP
            >>> for tasks in dc.computation.Future.as_completed(computation, timeout=10)
            ...     print(tasks.timing)   # doctest: +SKIP
            ...
            {'total_real_time': 17318, ... 'qpu_readout_time_per_sample': 123}
            {'total_real_time': 10816, ... 'qpu_readout_time_per_sample': 123}
            {'total_real_time': 26285, ... 'qpu_readout_time_per_sample': 123}
            ...
            >>> client.close()       # doctest: +SKIP

        """
        not_done = fs
        while not_done:
            done, not_done = Future.wait_multiple(not_done, min_done=1, timeout=timeout)
            if not done:
                raise TimeoutError
            for f in done:
                yield f

    def wait(self, timeout=None):
        """Wait for the solver to receive a response for a submitted problem.

        Blocking call that waits for a :class:`Future` object to complete.

        Args:
            timeout (float, optional, default=None):
                Maximum number of seconds to await completion. If None, waits
                indefinitely.

        Returns:
            Boolean: True if solver received a response.

        Examples:
            This example creates a solver using the local system's default
            D-Wave Cloud Client configuration file, submits a simple QUBO
            problem to a remote D-Wave resource for 100 samples, and tries
            waiting for 10 seconds for sampling to complete.

            >>> from dwave.cloud import Client
            >>> client = Client.from_config()         # doctest: +SKIP
            >>> solver = client.get_solver()          # doctest: +SKIP
            >>> u, v = next(iter(solver.edges))       # doctest: +SKIP
            >>> Q = {(u, u): -1, (u, v): 0, (v, u): 2, (v, v): -1}  # doctest: +SKIP
            >>> computation = solver.sample_qubo(Q, num_reads=100)   # doctest: +SKIP
            >>> computation.wait(timeout=10)    # doctest: +SKIP
            False
            >>> computation.remote_status        # doctest: +SKIP
            'IN_PROGRESS'
            >>> computation.wait(timeout=10)    # doctest: +SKIP
            True
            >>> computation.remote_status       # doctest: +SKIP
            'COMPLETED'
            >>> client.close()         # doctest: +SKIP
        """
        return self._results_ready_event.wait(timeout)

    def done(self):
        """Check whether the solver received a response for a submitted problem.

        Non-blocking call that checks whether the solver has received a response
        from the remote resource.

        Returns:
            Boolean: True if solver received a response.

        Examples:
            This example creates a solver using the local system's default
            D-Wave Cloud Client  configuration file, submits a simple QUBO
            problem to a remote D-Wave resource for 100 samples, and checks a
            couple of times whether sampling is completed.

            >>> from dwave.cloud import Client
            >>> client = Client.from_config()       # doctest: +SKIP
            >>> solver = client.get_solver()        # doctest: +SKIP
            >>> u, v = next(iter(solver.edges))     # doctest: +SKIP
            >>> Q = {(u, u): -1, (u, v): 0, (v, u): 2, (v, v): -1}   # doctest: +SKIP
            >>> computation = solver.sample_qubo(Q, num_reads=100)   # doctest: +SKIP
            >>> computation.done()  # doctest: +SKIP
            False
            >>> computation.done()   # doctest: +SKIP
            True
            >>> client.close()       # doctest: +SKIP
        """
        return self._results_ready_event.is_set()

    def cancel(self):
        """Try to cancel the problem corresponding to this result.

        Non-blocking call to the remote resource in a best-effort attempt to
        prevent execution of a problem.

        Examples:
            This example creates a solver using the local system's default
            D-Wave Cloud Client configuration file, submits a simple QUBO
            problem to a remote D-Wave resource for 100 samples, and tries
            (and in this case succeeds) to cancel it.

            >>> from dwave.cloud import Client
            >>> client = Client.from_config()         # doctest: +SKIP
            >>> solver = client.get_solver()          # doctest: +SKIP
            >>> u, v = next(iter(solver.edges))       # doctest: +SKIP
            >>> Q = {(u, u): -1, (u, v): 0, (v, u): 2, (v, v): -1}   # doctest: +SKIP
            >>> computation = solver.sample_qubo(Q, num_reads=100)   # doctest: +SKIP
            >>> computation.cancel()  # doctest: +SKIP
            >>> computation.done()   # doctest: +SKIP
            True
            >>> computation.remote_status    # doctest: +SKIP
            'CANCELLED'
            >>> client.close()      # doctest: +SKIP

        """
        # Don't need to cancel something already finished
        if self.done():
            return

        with self._single_cancel_lock:
            # Already done
            if self._cancel_requested:
                return

            # Set the cancel flag
            self._cancel_requested = True

            # The cancel request will be sent here, or by the solver when it
            # gets a status update for this problem (in the case where the id hasn't been set yet)
            if self.id is not None and not self._cancel_sent:
                self._cancel_sent = True
                self.solver.client._cancel(self.id, self)

    def wait_id(self, timeout=None):
        """Blocking id getter.

        Return the submitted problem ID, but unlike :meth:`.id`, block until the
        ID becomes known, or until `timeout` expires.

        Args:
            timeout (float, default=None):
                Timeout in seconds. By default, wait indefinitely for problem
                id to become known/available.

        Returns:
            str:
                Problem ID, as returned by SAPI.

        Raises:
            :exc:`concurrent.futures.TimeoutError`:
                When `timeout` exceeded, and problem id not ready.

        """
        if not self._id_ready_event.wait(timeout=timeout):
            raise TimeoutError("problem id not available yet")

        return self._id

    @property
    def id(self):
        """Simple non-blocking id getter for backward compat."""
        return self._id

    @id.setter
    def id(self, value):
        """Sets the problem ID, notifying the related event."""
        self._id = value

        # notify ID is set/ready
        if value is not None:
            self._id_ready_event.set()

    def result(self):
        """Results for a submitted job.

        Retrives raw result data in a :class:`Future` object that the solver
        submitted to a remote resource. First calls to access this data are
        blocking.

        Returns:
            dict: Results of the submitted job. Should be considered read-only.

        Note:
            Helper properties on :class:`Future` object are preferred to reading
            raw results, as they abstract away the differences in response
            between some solvers like. Available methods are: :meth:`samples`,
            :meth:`energies`, :meth:`variables`, :meth:`timing`,
            :meth:`problem_type`, :meth:`sampleset` (only if dimod package is
            installed).

        Warning:
            The dictionary returned by :meth:`result` depends on the solver
            used. Starting with version 0.7.0 we will not try to standardize
            them anymore, on client side. For QPU solvers, please replace
            `'samples'` with `'solutions'` and `'occurrences'` with
            `'num_occurrences'`. Better yet, use :meth:`Future.samples` and
            :meth:`Future.num_occurrences` instead.

        .. deprecated:: 0.8.0

            Alias keys ``samples`` and ``occurrences`` in the result dict are
            deprecated and will be removed in 0.12.0. We'll try to keep the
            result dict as close to raw data returned by SAPI as possible.
            Postprocessed data is available via
            :class:`~dwave.cloud.computation.Future` properties.

        .. versionchanged:: 0.8.0

            Instead of adding copies of ``solutions`` and ``num_occurrences``
            keys (as ``samples`` and ``occurrences``), we alias them using
            :class:`~dwave.cloud.utils.aliasdict`. Values are available under
            alias keys, but the keys themselves are not stored or visible.

        Examples:
            This example creates a solver using the local system's default
            D-Wave Cloud Client configuration file, submits a simple QUBO
            problem (representing a Boolean NOT gate by a penalty function)
            to a remote D-Wave resource for 5 samples, and prints part
            of the returned result (the relevant samples).

            >>> from dwave.cloud import Client
            >>> with Client.from_config() as client:  # doctest: +SKIP
            ...     solver = client.get_solver()
            ...     u, v = next(iter(solver.edges))
            ...     Q = {(u, u): -1, (u, v): 0, (v, u): 2, (v, v): -1}
            ...     computation = solver.sample_qubo(Q, num_reads=5)
            ...     for i in range(5):
            ...         result = computation.result()
            ...         print(result['solutions'][i][u], result['solutions'][i][v])
            ...
            ...
            (0, 1)
            (1, 0)
            (1, 0)
            (0, 1)
            (0, 1)

        """
        self._load_result()
        return self._result

    def exception(self):
        if self._exception is not None:
            raise self._exception

    @property
    def energies(self):
        """Energy buffer for the submitted job.

        First calls to access data of a :class:`Future` object are blocking;
        subsequent access to this property is non-blocking.

        Returns:
            list or NumPy matrix of doubles: Energies for each set of samples.

        Examples:
            This example creates a solver using the local system's default
            D-Wave Cloud Client configuration file, submits a random Ising
            problem (+1 or -1 values of linear and quadratic biases on all nodes
            and edges, respectively, of the solver's garph) to a remote D-Wave
            resource for 10 samples, and prints the returned energies.

            >>> import random
            >>> from dwave.cloud import Client
            >>> with Client.from_config() as client:  # doctest: +SKIP
            ...     solver = client.get_solver()
            ...     linear = {index: random.choice([-1, 1]) for index in solver.nodes}
            ...     quad = {key: random.choice([-1, 1]) for key in solver.undirected_edges}
            ...     computation = solver.sample_ising(linear, quad, num_reads=10)
            ...     print(computation.energies)
            ...
            [-3976.0, -3974.0, -3972.0, -3970.0, -3968.0, -3968.0, -3966.0,
             -3964.0, -3964.0, -3960.0]
        """

        # return energies from sampleset, if already constructed
        result = self.result()
        if 'sampleset' in result:
            return result['sampleset'].record.energy

        # fallback to energies from response
        return result['energies']

    @property
    def samples(self):
        """State buffer for the submitted job.

        First calls to access data of a :class:`Future` object are blocking;
        subsequent access to this property is non-blocking.

        Returns:
            list of lists or NumPy matrix: Samples on the nodes of solver's graph.

        Examples:
            This example creates a solver using the local system's default
            D-Wave Cloud Client configuration file, submits a simple QUBO
            problem (representing a Boolean NOT gate by a penalty function) to a
            remote D-Wave resource for 5 samples, and prints part of the
            returned result (the relevant samples).

            >>> from dwave.cloud import Client
            >>> with Client.from_config() as client:  # doctest: +SKIP
            ...     solver = client.get_solver()
            ...     u, v = next(iter(solver.edges))
            ...     Q = {(u, u): -1, (u, v): 0, (v, u): 2, (v, v): -1}
            ...     computation = solver.sample_qubo(Q, num_reads=5)
            ...     for i in range(5):
            ...         print(computation.samples[i][u], computation.samples[i][v])
            ...
            ...
            (1, 0)
            (0, 1)
            (0, 1)
            (1, 0)
            (0, 1)
        """
        # return samples from sampleset, if already constructed
        result = self.result()
        if 'sampleset' in result:
            return result['sampleset'].record.sample

        # fallback to samples from response
        return result['solutions']

    @property
    def variables(self):
        """List of active variables in response/answer."""

        result = self.result()

        if 'active_variables' in result:
            return result['active_variables']

        if 'sampleset' in result:
            return result['sampleset'].variables

        raise InvalidAPIResponseError("Active variables not present in the response")

    @property
    def num_occurrences(self):
        """Number of sample occurrences buffer for the submitted job.

        First calls to access data of a :class:`Future` object are blocking;
        subsequent access to this property is non-blocking.

        Returns:
            list or NumPy matrix of doubles: number of occurrences. When
            returned results are ordered in a histogram, ``num_occurrences``
            indicates the number of times a particular solution recurred.

        Examples:
            This example creates a solver using the local system's default
            D-Wave Cloud Client configuration file, submits a simple Ising
            problem with several ground states to a remote D-Wave resource for
            20 samples, and prints the returned results, which are ordered as a
            histogram. The problem's ground states tend to recur frequently,
            and so those solutions have ``num_occurrences`` greater than 1.

            >>> from dwave.cloud import Client
            >>> with Client.from_config() as client:  # doctest: +SKIP
            ...     solver = client.get_solver()
            ...     quad = {(16, 20): -1, (17, 20): 1, (16, 21): 1, (17, 21): 1}
            ...     computation = solver.sample_ising({}, quad, num_reads=500, answer_mode='histogram')
            ...     for i in range(len(computation.num_occurrences)):
            ...         print(computation.samples[i][16], computation.samples[i][17],
            ...               computation.samples[i][20], computation.samples[i][21],
            ...               ' --> ', computation.energies[i], computation.num_occurrences[i])
            ...
            (-1, 1, -1, -1, ' --> ', -2.0, 41)
            (-1, -1, -1, 1, ' --> ', -2.0, 53)
            (1, -1, 1, 1, ' --> ', -2.0, 55)
            (1, 1, -1, -1, ' --> ', -2.0, 52)
            (1, 1, 1, -1, ' --> ', -2.0, 60)
            (1, -1, 1, -1, ' --> ', -2.0, 196)
            (-1, 1, -1, 1, ' --> ', -2.0, 15)
            (-1, -1, 1, 1, ' --> ', -2.0, 28)

        """

        # return num_occurrences from sampleset, if already constructed
        result = self.result()
        if 'sampleset' in result:
            return result['sampleset'].record.num_occurrences

        # fallback to num_occurrences from response
        # (but `occurrences` data is not present if `answer_mode` was set to "raw")
        if 'num_occurrences' in result:
            return result['num_occurrences']
        elif self.return_matrix:
            return np.ones((len(result['solutions']),))
        else:
            return [1] * len(result['solutions'])

    def wait_sampleset(self):
        """Blocking sampleset getter."""

        # blocking result get
        result = self._load_result()

        # common problem info: id/label
        problem_info = dict(problem_id=self.id)
        if self.label is not None:
            problem_info.update(problem_label=self.label)

        # sapi returned sampleset directly
        if 'sampleset' in result:
            sampleset = result['sampleset']
            sampleset.info.update(problem_info)
            return sampleset

        # construct sampleset from available data
        try:
            import dimod
        except ImportError:
            raise RuntimeError("Can't construct SampleSet without dimod. "
                               "Re-install the library with 'bqm' support.")

        # filter inactive variables from samples
        variables = self.variables
        samples = [[sample[v] for v in variables] for sample in self.samples]

        # infer vartype from problem type
        # note: KeyError on unknown problem types. BQM/DQM should be handled above.
        vartype_from_problem_type = {'ising': 'SPIN', 'qubo': 'BINARY'}
        vartype = vartype_from_problem_type[self.problem_type]

        # for QPU jobs, info field is blank; add timing info
        info = dict(timing=self.timing)
        info.update(problem_info)

        sampleset = dimod.SampleSet.from_samples(
            (samples, variables), vartype=vartype,
            energy=self.energies, num_occurrences=self.num_occurrences,
            info=info, sort_labels=True)

        # the id is stored in the info field, but to be consistent with
        # the samplesets constructed .from_future, we add the method as well
        sampleset.wait_id = self.wait_id

        # this means that samplesets retrieved BEFORE this function are called
        # are not the same object as after, but it is a simpler implementation
        self._result['sampleset'] = self._sampleset = sampleset

        return sampleset

    @property
    def sampleset(self):
        """Return :class:`~dimod.SampleSet` representation of the results.

        Adds a `.wait_id` method to retrieve the id before the sampleset is
        resolved.
        """

        try:
            return self._sampleset
        except AttributeError:
            pass

        try:
            import dimod
        except ImportError:
            raise RuntimeError("Can't construct SampleSet without dimod. "
                               "Re-install the library with 'bqm' support.")

        self._sampleset = sampleset = dimod.SampleSet.from_future(
            self, lambda f: f.wait_sampleset())

        # propagate id to sampleset as well
        # note: this requires dimod>=0.8.21 (before that version SampleSet
        # had slots set which prevented dynamic addition of attributes).
        sampleset.wait_id = self.wait_id

        return sampleset

    @property
    def timing(self):
        """Timing information about a solver operation.

        Mapping from string keys to numeric values representing timing details
        for a submitted job as returned from the remote resource. Keys are
        dependant on the particular solver.

        First calls to access data of a :class:`Future` object are blocking;
        subsequent access to this property is non-blocking.

        Returns:
            dict:
                Mapping from string keys to numeric values representing timing
                information.

        Examples:
            This example creates a client using the local system's default
            D-Wave Cloud Client configuration file, which is configured to
            access a D-Wave 2000Q QPU, submits a simple :term:`Ising` problem
            (opposite linear biases on two coupled qubits) for 5 samples, and
            prints timing information for the job.

            >>> from dwave.cloud import Client
            >>> with Client.from_config() as client:      # doctest: +SKIP
            ...     solver = client.get_solver()
            ...     u, v = next(iter(solver.edges))
            ...     computation = solver.sample_ising({u: -1, v: 1},{}, num_reads=5)
            ...     print(computation.timing)
            ...
            {'total_real_time': 10961, 'anneal_time_per_run': 20, ...}

        """
        return self.result().get('timing', {})

    @property
    def problem_type(self):
        """Submitted problem type for this computation, as returned by the
        solver API. Typical values are 'ising' and 'qubo'.
        """
        return self.result()['problem_type']

    def __getitem__(self, key):
        """Provide a simple results item getter. Blocks if future is unresolved.

        Args:
            key: keywords for result fields.
        """
        self._load_result()
        if key not in self._result:
            raise KeyError('{} is not a property of response object'.format(key))
        return self._result[key]

    def _load_result(self):
        """Get the result, waiting and decoding as needed."""
        if self._result is None:
            # Wait for the query response
            self.wait(timeout=None)

            # Check for other error conditions
            if self._exception is not None:
                raise self._exception

            # If someone else took care of this while we were waiting
            if self._result is not None:
                return self._result

            # Prepare results from the response
            self._decode()
            self._alias_result()

        return self._result

    def _patch_offset(self):
        # XXX: This is a temporary fix, until SAPI starts returning the offset
        # in answer (for structured solvers only).
        # It will patch `self._message` to include the offset as set in
        # `self._offset`, but only if SAPI answer does not contain offset already.
        msg = self._message
        fmt = msg.get('answer', {}).get('format')
        if fmt == 'qp':
            if 'offset' not in msg['answer']:
                msg['answer']['offset'] = self._offset

    def _decode(self):
        """Decode answer data from the response."""
        start = time.time()
        self._patch_offset()
        self._result = self.solver.decode_response(self._message)
        self.parse_time = time.time() - start
        return self._result

    def _alias_result(self):
        """Alias `solutions` and `num_occurrences`.

        Deprecated in version 0.8.0.

        Scheduled for removal in 0.12.0.
        """
        if not self._result:
            return

        msg = "'{}' alias has been deprecated in favor of '{}'"
        samples_msg = msg.format('samples', 'solutions')
        occurrences_msg = msg.format('occurrences', 'num_occurrences')

        aliases = dict(
            samples=deprecated(samples_msg)(itemgetter('solutions')),
            occurrences=deprecated(occurrences_msg)(itemgetter('num_occurrences')))

        self._result = aliasdict(self._result)
        self._result.alias(aliases)

        return self._result



# Copyright 2020 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

try:
    import dimod
except ImportError:
    dimod = None

from dwave.cloud.client import Client
from dwave.cloud.solver import StructuredSolver, UnstructuredSolver
from dwave.cloud.events import add_handler, dispatches_events
from dwave.cloud.concurrency import Present


class TestEventDispatch(unittest.TestCase):

    def setUp(self):
        # mock client
        self.client = Client(endpoint='e', token='t', solver=dict(name__contains='test'))
        self.client._fetch_solvers = lambda **kw: self.solvers
        self.client._submit = lambda *pa, **kw: None
        self.client.upload_problem_encoded = lambda *pa, **kw: Present(result=mock_problem_id)

        # mock solvers
        self.structured_solver = StructuredSolver(client=self.client, data={
            "properties": {
                "supported_problem_types": ["qubo", "ising"],
                "qubits": [0, 1, 2],
                "couplers": [[0, 1], [0, 2], [1, 2]],
                "num_qubits": 3,
                "num_reads_range": [0, 100],
                "parameters": {
                    "num_reads": "Number of samples to return.",
                    "postprocess": "either 'sampling' or 'optimization'"
                },
                "topology": {
                    "type": "chimera",
                    "shape": [16, 16, 4]
                },
                "category": "qpu",
                "tags": ["lower_noise"]
            },
            "id": "test-qpu-solver",
            "description": "A test solver 1",
            "status": "online"
        })
        self.unstructured_solver = UnstructuredSolver(client=self.client, data={
            "properties": {
                "supported_problem_types": ["bqm"],
                "parameters": {"num_reads": "Number of samples to return."},
                "category": "hybrid",
            },
            "id": "test-unstructured-solver",
            "description": "A test unstructured solver"
        })
        self.solvers = [self.structured_solver]
        # we can't use unstructured solvers without dimod installed,
        # so don't even try testing it
        if dimod:
            self.solvers.append(self.unstructured_solver)

        # reset all event handlers
        from dwave.cloud.events import _client_event_hooks_registry as reg
        reg.update({k: [] for k in reg})

    def test_validation(self):
        """Event name and handler are validated."""

        with self.assertRaises(ValueError):
            add_handler('invalid_event_name', lambda: None)
        with self.assertRaises(TypeError):
            add_handler('before_client_init', None)

    def test_client_init(self):
        """Before/After client init events are dispatched with correct signatures."""

        # setup event handlers
        memo = {}
        def handler(event, **data):
            memo[event] = data

        add_handler('before_client_init', handler)
        add_handler('after_client_init', handler)

        # client init
        client = Client(endpoint='endpoint', token='token', unknown='unknown')

        # test entry values
        before = memo['before_client_init']
        self.assertEqual(before['obj'], client)
        self.assertIn('kwargs', before['args'])
        self.assertIn('endpoint', before['args']['kwargs'])
        self.assertIn('token', before['args']['kwargs'])
        self.assertEqual(before['args']['kwargs']['token'], 'token')
        self.assertEqual(before['args']['kwargs']['unknown'], 'unknown')

        # test exit values
        after = memo['after_client_init']
        self.assertEqual(after['obj'], client)
        self.assertEqual(after['args']['kwargs']['token'], 'token')
        self.assertEqual(after['args']['kwargs']['unknown'], 'unknown')
        self.assertEqual(after['args']['kwargs']['endpoint'], 'endpoint')
        self.assertEqual(after['return_value'], None)

    def test_get_solvers(self):
        """Before/After get_solvers events are dispatched with correct signatures."""

        # setup event handlers
        memo = {}
        def handler(event, **data):
            memo[event] = data

        add_handler('before_get_solvers', handler)
        add_handler('after_get_solvers', handler)

        # get solver(s)
        self.client.get_solver()

        # test entry values
        before = memo['before_get_solvers']
        self.assertEqual(before['obj'], self.client)
        self.assertIn('refresh', before['args'])
        self.assertIn('filters', before['args'])
        self.assertIn('name__contains', before['args']['filters'])

        # test exit values
        after = memo['after_get_solvers']
        self.assertEqual(after['obj'], self.client)
        self.assertIn('name__contains', after['args']['filters'])
        self.assertEqual(after['return_value'], self.solvers)

    def subtest_sample(self, solver):
        # setup event handlers
        memo = {}
        def handler(event, **data):
            memo[event] = data

        add_handler('before_sample', handler)
        add_handler('after_sample', handler)

        # sample
        lin = {0: 1}
        quad = {(0, 1): 1}
        offset = 2
        params = dict(num_reads=100)
        future = solver.sample_ising(lin, quad, offset, **params)

        # test entry values
        before = memo['before_sample']
        if solver.qpu:
            args = dict(type_='ising', linear=lin, quadratic=quad,
                        offset=offset, params=params,
                        undirected_biases=False, label=None)
        elif solver.hybrid:
            if not dimod:
                self.skipTest("dimod not installed")
            bqm = dimod.BQM.from_ising(lin, quad, offset)
            args = dict(problem=bqm, problem_type=None, label=None, params=params)

        self.assertEqual(before['obj'], solver)
        self.assertDictEqual(before['args'], args)

        # test exit values
        after = memo['after_sample']
        self.assertEqual(after['obj'], solver)
        self.assertDictEqual(after['args'], args)
        self.assertEqual(after['return_value'], future)

    def test_sample(self):
        """Before/After solver sample events are dispatched with correct signatures."""

        for solver in self.solvers:
            with self.subTest("solver=%r" % solver.id):
                self.subtest_sample(solver)


class TestEventDispatchDecorator(unittest.TestCase):

    def setUp(self):
        # reset all event handlers
        from dwave.cloud.events import _client_event_hooks_registry as reg
        reg.update({k: [] for k in reg})

    def test_decorator(self):
        """Decorator adds on-entry and on-exit event calls, with correct args."""

        class MockSampler:
            @dispatches_events('sample')
            def mock_sample(self, h, J, offset=0, fail=False, **kwargs):
                if fail:
                    raise ValueError
                return offset + 1

        mock_object = MockSampler()
        h = [1, 1]
        J = {(0, 1): 1}

        def before(name, obj, args):
            self.assertEqual(obj, mock_object)
            args.pop('fail')
            self.assertEqual(args, dict(h=h, J=J, offset=0, kwargs={}))

        def after(name, obj, args, return_value=None, exception=None):
            self.assertEqual(obj, mock_object)
            fail = args.pop('fail')
            self.assertEqual(args, dict(h=h, J=J, offset=0, kwargs={}))
            if fail:
                self.assertIsInstance(exception, ValueError)
            else:
                self.assertEqual(return_value, 1)

        add_handler('before_sample', before)
        add_handler('before_sample', after)

        mock_object.mock_sample(h, J)
        with self.assertRaises(ValueError):
            mock_object.mock_sample(h, J, fail=True)



# Copyright 2019 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Concurrency utilities."""

import sys
import functools
import concurrent.futures
import queue

__all__ = ['PriorityThreadPoolExecutor']


@functools.total_ordering
class _PriorityOrderedItem(object):
    """Generic priority queue item with ordering defined according to the
    priority attribute.
    """

    def __init__(self, item, priority=None):
        # by default, None < Any
        if priority is None:
            if item is None:
                priority = -sys.maxsize
            else:
                priority = sys.maxsize

        self.item = item
        self.priority = priority

    def __lt__(self, other):
        return self.priority < other.priority

    def __eq__(self, other):
        return self.priority == other.priority


class _PrioritizedWorkItem(_PriorityOrderedItem):
    """Extension of :class:`_PriorityOrderedItem` that handles execution
    priority passed in `kwargs` of :class:`concurrent.futures.thread._WorkItem`
    items.
    """

    def __init__(self, item):
        if not isinstance(item, concurrent.futures.thread._WorkItem):
            raise TypeError("concurrent.futures.thread._WorkItem expected")

        # copy constructor
        if isinstance(item, _PrioritizedWorkItem):
            priority = item.priority
        else:
            priority = item.kwargs.pop('priority', sys.maxsize)

        super().__init__(item, priority)


class _PrioritizingQueue(queue.PriorityQueue):
    """Re-pack :class:`concurrent.futures.thread._WorkItem` (on queue put) into
    :class:`_PrioritizedWorkItem` subclass that is ordered according to
    priority.
    """

    def put(self, item, *args, **kwargs):
        # unpack item, extract priority
        if isinstance(item, concurrent.futures.thread._WorkItem):
            item = _PrioritizedWorkItem(item)
        else:
            item = _PriorityOrderedItem(item)

        super().put(item, *args, **kwargs)

    def get(self, *args, **kwargs):
        prioritized_item = super().get(*args, **kwargs)
        return prioritized_item.item


class PriorityThreadPoolExecutor(concurrent.futures.ThreadPoolExecutor):
    """Add support for task priority to the standard
    :class:`concurrent.futures.ThreadPoolExecutor` FIFO-queue based
    :class:`concurrent.futures.Executor` implementation.

    Interface is identical to :class:`concurrent.futures.ThreadPoolExecutor`,
    except the `.submit()` which now accepts optional `priority` keyword
    argument::

        def submit(self, fn, *args, priority=sys.maxsize, **kwargs):
            ...

    Note: if `priority` is omitted, the behavior is identical to that of the
    superclass, `ThreadPoolExecutor`.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._work_queue = _PrioritizingQueue()


class Present(concurrent.futures.Future):
    """Already resolved :class:`~concurrent.futures.Future` object.

    Users should treat this class as just another
    :class:`~concurrent.futures.Future`, the difference being an implementation
    detail: :class:`Present` is "resolved" at construction time.
    """

    def __init__(self, result=None, exception=None):
        super().__init__()
        if result is not None:
            self.set_result(result)
        elif exception is not None:
            self.set_exception(exception)
        else:
            raise ValueError("can't provide both 'result' and 'exception'")



# Copyright 2021 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import uuid
import itertools
import random
from typing import List, Tuple, Optional

__all__ = [
    'solver_configuration_data',
    'structured_solver_data', 'qpu_clique_solver_data',
    'qpu_chimera_solver_data', 'qpu_pegasus_solver_data', 'qpu_problem_timing_data',
    'unstructured_solver_data', 'hybrid_bqm_solver_data', 'hybrid_dqm_solver_data',
]


def solver_configuration_data(id: str = None,
                              status: str = None,
                              description: str = None,
                              properties: dict = None,
                              avg_load: float = None) -> dict:
    """Return data dict describing a solver, as returned by SAPI."""

    if id is None:
        id = str(uuid.uuid4())

    if status is None:
        status = "ONLINE"

    if description is None:
        description = "A mock solver"

    if properties is None:
        properties = {}

    if avg_load is None:
        avg_load = random.random()

    return {
        "id": id,
        "status": status,
        "description": description,
        "properties": properties,
        "avg_load": avg_load,
    }


def structured_solver_data(id: str = None,
                           status: str = None,
                           avg_load: float = None,
                           description: str = None,
                           properties: dict = None,
                           # specific properties for convenience, override `properties` dict
                           qubits: List[int] = None,
                           couplers: List[List[int]] = None,
                           **kwargs) -> dict:
    """Return data dict describing a structured solver. By default, solver has
    only one qubit.
    """

    if properties is None:
        # TODO: add more default properties?
        properties = {
            "category": "qpu",
            "tags": ["lower_noise"],
            "topology": {"type": "clique"},
            "supported_problem_types": ["qubo", "ising"],
            "h_range": [-2.0, 2.0],
            "j_range": [-1.0, 1.0],
            "num_reads_range": [1, 10000],
            "annealing_time_range": [1.0, 2000.0],
            "extended_j_range": [-2.0, 1.0],
            "quota_conversion_rate": 1,
            "parameters": {
                "anneal_offsets": "A list of anneal offsets for each working qubit.",
                "anneal_schedule": "A piecewise linear annealing schedule.",
                "label": "Problem label.",
                "num_reads": "Number of samples to return."
            }
        }

    if qubits is None:
        qubits = [0]

    if couplers is None:
        couplers = []

    properties.update({
        "qubits": qubits,
        "couplers": couplers,
        "num_qubits": len(qubits)
    })

    properties.update(kwargs)

    return solver_configuration_data(id=id,
                                     status=status,
                                     description=description,
                                     properties=properties,
                                     avg_load=avg_load)


def qpu_clique_solver_data(size: int, **kwargs) -> dict:
    """Mock QPU solver data with a clique of custom size as topology.

    Args:
        size:
            Clique size. Qubits are sequential integers in range [0, size).
        **kwargs:
            Solver properties passed down to :meth:`.structured_solver_data`.
    """

    qubits = list(range(size))
    couplers = list(itertools.combinations(range(len(qubits)), 2))

    params = dict(
        id=f"clique_{size}q_mock",
        description=f"A {size}-qubit mock QPU solver with clique topology",
        qubits=qubits,
        couplers=couplers,
        topology={"type": "clique", "shape": [size]},
    )
    params.update(**kwargs)

    return structured_solver_data(**params)


def qpu_chimera_solver_data(m: int,
                            n: Optional[int] = None,
                            t: Optional[int] = None,
                            **kwargs) -> dict:
    """Mock QPU solver data with a custom-sized Chimera topology.

    Args:
        m:
            Number of rows in the Chimera lattice. See
            :func:`~dwave_networkx.generators.chimera_graph` for details.
        n:
            Number of columns in the Chimera lattice.
        t:
            Size of the shore within each Chimera tile.
        **kwargs:
            Solver properties passed down to :meth:`.structured_solver_data`.
    """
    try:
        import dwave_networkx as dnx
    except ImportError:     # pragma: no cover
        raise RuntimeError("Can't generate Chimera graph without dwave-networkx. "
                           "Install with 'dwave-cloud-client[mocks]'.")

    graph = dnx.chimera_graph(m, n, t)
    # we need generated graph's values, so we can set topology.shape
    m = graph.graph['rows']
    n = graph.graph['columns']
    t = graph.graph['tile']
    qubits = list(graph.nodes)
    couplers = list(graph.edges)
    num_qubits = len(qubits)

    params = dict(
        id=f"chimera_{num_qubits}q_mock",
        description=f"A {num_qubits}-qubit mock QPU solver with chimera topology",
        qubits=qubits,
        couplers=couplers,
        topology={"type": "chimera", "shape": [m, n, t]},
    )
    params.update(**kwargs)

    return structured_solver_data(**params)


def qpu_pegasus_solver_data(m: int,
                            fabric_only: bool = True,
                            **kwargs) -> dict:
    """Mock QPU solver data with a custom-sized Pegasus topology.

    Args:
        m:
            Size parameter for the Pegasus lattice.
        fabric_only:
            Use only nodes from the largest Pegasus graph component. See
            :func:`~dwave_networkx.generators.pegasus_graph` for details.
        **kwargs:
            Solver properties passed down to :meth:`.structured_solver_data`.

    Note:
        By default, with ``fabric_only=True``, only a subset of Pegasus graph
        is used (fabric qubits only), hence num_active_qubits will be less than
        num_qubits.
    """
    try:
        import dwave_networkx as dnx
    except ImportError:     # pragma: no cover
        raise RuntimeError("Can't generate Pegasus graph without dwave-networkx. "
                           "Install with 'dwave-cloud-client[mocks]'.")

    graph = dnx.pegasus_graph(m, fabric_only=fabric_only)
    qubits = list(graph.nodes)
    couplers = list(graph.edges)
    num_qubits = 24 * m * (m-1)     # includes non-fabric qubits

    params = dict(
        id=f"pegasus_{num_qubits}q_mock",
        description=f"A {num_qubits}-qubit mock QPU solver with pegasus topology",
        qubits=qubits,
        couplers=couplers,
        topology={"type": "pegasus", "shape": [m]},
        num_qubits=num_qubits,
    )
    params.update(**kwargs)

    return structured_solver_data(**params)

def qpu_problem_timing_data(qpu: str = 'advantage') -> dict:
    """Mock QPU solver proprty problem_timing_data.

    Args:
        qpu:
            Type of QPU that provided timing data, as it was in
            August 2022 on a particular system, used here. Currently supported
            values are: ``advantage`` and ``2000q``.

    """

    timing_data_advantage41_2022_8 = {'version': '1.0.0',
        'typical_programming_time': 14072.88,
        'reverse_annealing_with_reinit_prog_time_delta': 0.0,
        'reverse_annealing_without_reinit_prog_time_delta': 5.55,
        'default_programming_thermalization': 1000.0,
        'default_annealing_time': 20.0,
        'readout_time_model': 'pwl_log_log',
        'readout_time_model_parameters': [0.0, 0.7699665876947938, 1.7242758696010096,
                2.711975459489206, 3.1639057672764026, 3.750276915153992, 1.539131714800995,
                1.8726623164229292, 2.125631787097315, 2.332672340068556, 2.371606651233025,
                2.3716219271760215],
        'qpu_delay_time_per_sample': 20.54,
        'reverse_annealing_with_reinit_delay_time_delta': -4.5,
        'reverse_annealing_without_reinit_delay_time_delta': -1.5,
        'default_readout_thermalization': 0.0,
        'decorrelation_max_nominal_anneal_time': 2000.0,
        'decorrelation_time_range': [500.0, 10000.0]}

    timing_data_2000q6_2022_8 = {'version': '1.0.0',
        'typical_programming_time': 10536.81,
        'reverse_annealing_with_reinit_prog_time_delta': 0.0,
        'reverse_annealing_without_reinit_prog_time_delta': 359.58,
        'default_programming_thermalization': 1000.0,
        'default_annealing_time': 20.0,
        'readout_time_model': 'pwl_log_log',
        'readout_time_model_parameters': [0.0, 3.30984300471607, 2.2975416678181597,
            2.2975416678181597],
        'qpu_delay_time_per_sample': 20.54,
        'reverse_annealing_with_reinit_delay_time_delta': 586.38,
        'reverse_annealing_without_reinit_delay_time_delta': -5.0,
        'default_readout_thermalization': 0.0,
        'decorrelation_max_nominal_anneal_time': 2000.0,
        'decorrelation_time_range': [500.0, 10000.0]}

    name_dict = {'advantage': timing_data_advantage41_2022_8,
                 '2000q': timing_data_2000q6_2022_8}

    return name_dict[qpu]

def unstructured_solver_data(id: str = None,
                             status: str = None,
                             avg_load: float = None,
                             description: str = None,
                             properties: dict = None,
                             # specific properties for convenience, override `properties` dict
                             supported_problem_types: List[str] = None,
                             **kwargs) -> dict:
    """Return data dict describing an unstructured solver."""

    if properties is None:
        properties = {
            "category": "hybrid",
            "maximum_number_of_variables": 65536,
            "maximum_time_limit_hrs": 2.0,
            "minimum_time_limit": [
                [1, 1.0], [1024, 2.0], [4096, 4.0], [8192, 16.0]
            ],
            "quota_conversion_rate": 20,
            "parameters": {
                "time_limit": "Hybrid solver execution time limit."
            }
        }

    if supported_problem_types is None:
        supported_problem_types = ["bqm"]

    properties.update({
        "supported_problem_types": supported_problem_types
    })

    properties.update(kwargs)

    return solver_configuration_data(id=id,
                                     status=status,
                                     description=description,
                                     properties=properties,
                                     avg_load=avg_load)


def hybrid_bqm_solver_data(**kwargs) -> dict:
    params = dict(
        id="hybrid_bqm_solver",
        description="Hybrid unstructured BQM mock solver",
        supported_problem_types=["bqm"]
    )
    params.update(**kwargs)
    return unstructured_solver_data(**params)


def hybrid_dqm_solver_data(**kwargs) -> dict:
    params = dict(
        id="hybrid_dqm_solver",
        description="Hybrid unstructured DQM mock solver",
        supported_problem_types=["dqm"]
    )
    params.update(**kwargs)
    return unstructured_solver_data(**params)



# Copyright 2017 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
An implementation of the REST client for D-Wave Solver API (SAPI) service.

SAPI servers provide authentication, queuing, and scheduling services,
and provide a network interface to :term:`solver`\ s. This API enables you submit
a binary quadratic (:term:`Ising` or :term:`QUBO`) model
and receive samples from a distribution over the model as defined by a
selected solver.

SAPI server workflow is roughly as follows:

 1. Submitted problems enter an input queue. Each user has an input queue per solver.
 2. Drawing from all input queues for a solver, problems are scheduled.
 3. Results are cached for retrieval by the client.

"""

from dwave.cloud.client import Client as BaseClient
from dwave.cloud.solver import StructuredSolver as Solver

__all__ = ['Client']


class Client(BaseClient):
    """D-Wave Solver API client specialized to work only with QPU solvers.

    This class can be instantiated explicitly, or via (base) Client's factory
    method, :meth:`~dwave.cloud.client.Client.from_config` by supplying
    ``"qpu"`` for ``client``.

    Examples:
        This example explicitly instantiates a :class:`dwave.cloud.qpu.Client`.
        :meth:`~dwave.cloud.client.Client.get_solver` is guaranteed to return a
        QPU solver.

        .. code-block:: python

            from dwave.cloud.qpu import Client

            with Client(token='...') as client:
                solver = client.get_solver()
                response = solver.sample_ising(...)

        The following example instantiates a QPU client indirectly. Again,
        :meth:`~dwave.cloud.client.Client.get_solver`/
        :meth:`~dwave.cloud.client.Client.get_solvers` are guaranteed to return
        only QPU solver(s).

        .. code-block:: python

            from dwave.cloud import Client

            with Client.from_config(client='qpu') as client:
                solver = client.get_solver()
                response = solver.sample_ising(...)

    """

    @staticmethod
    def is_solver_handled(solver):
        """Determine if the specified solver should be handled by this client.

        This predicate function (used from the base class) allows only remote
        QPU solvers.
        """
        return solver and solver.qpu



# Copyright 2020 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Early prototype of an event subsystem that allows external actors to hook into
data processing pipeline and access otherwise private data for closer
inspection.
"""

import logging
import inspect
from functools import wraps

__all__ = ['add_handler']

logger = logging.getLogger(__name__)


# package-global event hooks registry
# Dict[event_name: str, handler_functions: list]
_client_event_hooks_registry = {
    'before_client_init': [],
    'after_client_init': [],
    'before_get_solvers': [],
    'after_get_solvers': [],
    'before_sample': [],
    'after_sample': [],
}


# TODO: rewrite as decorator that automatically captures function input/output
def add_handler(name, handler):
    """Register a `handler` function to be called on event `name`.

    Handler signatures are::

        def before_event_handler(event_name, obj, args):
            # called just before `obj.method(**args)` executes

        def after_event_handler(event_name, obj, args, return_value):
            # function succeeded with `return_value`

        def after_event_handler(event_name, obj, args, exception):
            # function failed with `exception` raised
            # after event handler invocation, exception is re-raised

    """

    if name not in _client_event_hooks_registry:
        raise ValueError('invalid hook name')
    if not callable(handler):
        raise TypeError('callable handler required')

    _client_event_hooks_registry[name].append(handler)


def dispatch_event(name, *args, **kwargs):
    """Call the complete chain of event handlers attached to event `name`."""

    logger.trace("dispatch_event(%r, *%r, **%r)", name, args, kwargs)

    if name not in _client_event_hooks_registry:
        raise ValueError('invalid event name')

    for handler in _client_event_hooks_registry[name]:
        try:
            handler(name, *args, **kwargs)
        except Exception as e:
            logger.debug("Exception in {!r} event handler {!r}: {!r}".format(
                name, handler, e))


class dispatches_events:
    """Decorate function to :func:`.dispatch_event` on entry and exit."""

    def __init__(self, basename):
        self.before_eventname = 'before_' + basename
        self.after_eventname = 'after_' + basename

    def __call__(self, fn):
        if not callable(fn):
            raise TypeError("decorated object must be callable")

        @wraps(fn)
        def wrapped(*pargs, **kwargs):
            sig = inspect.signature(fn)
            bound = sig.bind(*pargs, **kwargs)
            bound.apply_defaults()
            args = bound.arguments
            obj = args.pop('self', None)

            dispatch_event(self.before_eventname, obj=obj, args=args)
            try:
                rval = fn(*pargs, **kwargs)
            except Exception as exc:
                dispatch_event(self.after_eventname, obj=obj, args=args, exception=exc)
                raise
            else:
                dispatch_event(self.after_eventname, obj=obj, args=args, return_value=rval)
                return rval

        return wrapped



# Copyright 2021 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from urllib.parse import urljoin

import requests_mock

from dwave.cloud.api.resources import Regions
from dwave.cloud.api import exceptions, models

from tests import config


class TestMockRegions(unittest.TestCase):
    """Test request formation and response parsing (including error handling)
    works correctly for all :class:`dwave.cloud.api.resources.Solvers` methods.
    """

    endpoint = 'http://test.com/path/'

    def setUp(self):
        self.mocker = requests_mock.Mocker()

        region1_data = {"code": "A", "name": "Region A", "endpoint": urljoin(self.endpoint, 'regions/A')}
        region1_code = region1_data['code']
        region1_uri = region1_data['endpoint']

        region2_data = {"code": "B", "name": "Region B", "endpoint": urljoin(self.endpoint, 'regions/B')}
        region2_code = region2_data['code']
        region2_uri = region2_data['endpoint']

        all_region_data = [region1_data, region2_data]
        all_region_data_uri = urljoin(self.endpoint, 'regions/')
        self.region_codes = [region1_code, region2_code]

        self.mocker.get(requests_mock.ANY, status_code=404)

        self.mocker.get(region1_uri, json=region1_data)
        self.mocker.get(region2_uri, json=region2_data)
        self.mocker.get(all_region_data_uri, json=all_region_data)

        self.mocker.start()

    def tearDown(self):
        self.mocker.stop()

    def test_list_regions(self):
        resource = Regions(endpoint=self.endpoint)
        regions = resource.list_regions()

        self.assertEqual(len(regions), len(self.region_codes))
        self.assertEqual(regions[0].code, self.region_codes[0])
        self.assertEqual(regions[1].code, self.region_codes[1])

    def test_get_region(self):
        code = self.region_codes[0]

        regions = Regions(endpoint=self.endpoint)
        region = regions.get_region(code)

        self.assertEqual(region.code, code)
        self.assertEqual(region.endpoint, urljoin(self.endpoint, f"regions/{code}"))

    def test_nonexisting_region(self):
        resource = Regions(endpoint=self.endpoint)
        with self.assertRaises(exceptions.ResourceNotFoundError):
            resource.get_region('non-existing-region')


@unittest.skipUnless(config, "API access not configured.")
class TestCloudRegions(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.api = Regions()

    @classmethod
    def tearDownClass(cls):
        cls.api.close()

    def test_list_solvers(self):
        regions = self.api.list_regions()

        self.assertIsInstance(regions, list)
        self.assertGreater(len(regions), 0)

        for region in regions:
            self.assertIsInstance(region, models.Region)

    def test_get_region(self):
        # don't assume solver availability, instead try fetching one from the list
        regions = self.api.list_regions()
        region_code = regions.pop().code

        region = self.api.get_region(region_code)
        self.assertIsInstance(region, models.Region)
        self.assertEqual(region.code, region_code)

    def test_nonexisting_region(self):
        with self.assertRaises(exceptions.ResourceNotFoundError):
            self.api.get_region('non-existing-region')
